# Support branch detection within clusters.
import numpy as np

from sklearn.base import BaseEstimator, ClusterMixin
from sklearn.neighbors import KDTree, BallTree
from scipy.sparse import coo_array
from scipy.sparse.csgraph import minimum_spanning_tree
from joblib import Memory
from joblib import Parallel, delayed
from joblib.parallel import cpu_count
from .dist_metrics import DistanceMetric
from ._hdbscan_linkage import label
from .plots import CondensedTree, SingleLinkageTree, ApproximationGraph
from .prediction import approximate_predict
from ._hdbscan_tree import (
    get_branches,
    condense_tree,
    recurse_leaf_dfs,
    compute_stability,
    simplify_branch_hierarchy,
)


class BranchDetectionData(object):
    """Input data for branch detection functionality.

    Recreates and caches internal data structures from the clustering stage.

    Parameters
    ----------

    data : array (n_samples, n_features)
        The original data set that was clustered.

    labels : array (n_samples)
        The cluster labels for every point in the data set.

    min_samples : int
        The min_samples value used in clustering.

    tree_type : string, optional
        Which type of space tree to use for core distance computation.
        One of:
            * ``kdtree``
            * ``balltree``

    metric : string, optional
        The metric used to determine distance for the clustering.
        This is the metric that will be used for the space tree to determine
        core distances etc.

    **kwargs :
        Any further arguments to the metric.

    Attributes
    ----------

    all_finite : bool
        Whether the data set contains any infinite or NaN values.

    finite_index : array (n_samples)
        The indices of the finite data points in the original data set.

    internal_to_raw : dict
        A mapping from the finite data set indices to the original data set.

    tree : KDTree or BallTree
        A space partitioning tree that can be queried for nearest neighbors if
        the metric is supported by a KDTree or BallTree.

    neighbors : array (n_samples, min_samples)
        The nearest neighbor for every non-noise point in the original data set.

    core_distances : array (n_samples)
        The core distance for every non-noise point in the original data set.

    dist_metric : callable
        Accelerated distance metric function.
    """

    _tree_type_map = {"kdtree": KDTree, "balltree": BallTree}

    def __init__(
        self,
        data,
        all_finite,
        finite_index,
        labels,
        min_samples,
        tree_type="kdtree",
        metric="euclidean",
        **kwargs,
    ):
        # Select finite data points
        self.all_finite = all_finite
        self.finite_index = finite_index
        clean_data = data.astype(np.float64)
        if not all_finite:
            labels = labels[finite_index]
            clean_data = clean_data[finite_index]
            self.internal_to_raw = {
                x: y for x, y in zip(range(len(finite_index)), finite_index)
            }
        else:
            self.internal_to_raw = None

        # Construct tree
        self.tree = self._tree_type_map[tree_type](clean_data, metric=metric, **kwargs)
        self.dist_metric = DistanceMetric.get_metric(metric, **kwargs)

        # Allocate to maintain data point indices
        self.core_distances = np.full(clean_data.shape[0], np.nan)
        self.neighbors = np.full((clean_data.shape[0], min_samples), -1, dtype=np.int64)

        # Find neighbours for non-noise points
        noise_mask = labels != -1
        if noise_mask.any():
            distances, self.neighbors[noise_mask, :] = self.tree.query(
                clean_data[noise_mask], k=min_samples
            )
            self.core_distances[noise_mask] = distances[:, -1]


def detect_branches_in_clusters(
    clusterer,
    min_branch_size=None,
    allow_single_branch=False,
    branch_detection_method="full",
    branch_selection_method="eom",
    branch_selection_persistence=0.0,
    max_branch_size=0,
    label_sides_as_branches=False,
):
    """
    Performs a flare-detection post-processing step to detect branches within
    clusters [1]_.

    For each cluster, a graph is constructed connecting the data points based on
    their mutual reachability distances. Each edge is given a centrality value
    based on how far it lies from the cluster's center. Then, the edges are
    clustered as if that centrality was a distance, progressively removing the
    'center' of each cluster and seeing how many branches remain.

    Parameters
    ----------

    clusterer : hdbscan.HDBSCAN
        The clusterer object that has been fit to the data with branch detection
        data generated.

    min_branch_size : int, optional (default=None)
        The minimum number of samples in a group for that group to be
        considered a branch; groupings smaller than this size will seen as
        points falling out of a branch. Defaults to the clusterer's min_cluster_size.

    allow_single_branch : bool, optional (default=False)
        Analogous to ``allow_single_cluster``.

    branch_detection_method : str, optional (default=``full``)
        Deteremines which graph is conctructed to detect branches with. Valid
        values are, ordered by increasing computation cost and decreasing
        sensitivity to noise:
        - ``core``: Contains the edges that connect each point to all other
          points within a mutual reachability distance lower than or equal to
          the point's core distance. This is the cluster's subgraph of the
          k-NN graph over the entire data set (with k = ``min_samples``).
        - ``full``: Contains all edges between points in each cluster with a
          mutual reachability distance lower than or equal to the distance of
          the most-distance point in each cluster. These graphs represent the
          0-dimensional simplicial complex of each cluster at the first point in
          the filtration where they contain all their points.

    branch_selection_method : str, optional (default='eom')
        The method used to select branches from the cluster's condensed tree.
        The standard approach for FLASC is to use the ``eom`` approach.
        Options are:
          * ``eom``
          * ``leaf``

    branch_selection_persistence: float, optional (default=0.0)
        An eccentricity persistence threshold. Branches with a persistence below
        this value will be merged. See [3]_ for more information. Note that this
        should not be used if we want to predict the cluster labels for new
        points in future (e.g. using approximate_predict), as the
        :func:`~flasc.prediction.approximate_predict` function is not aware of
        this argument.

    max_branch_size : int, optional (default=0)
        A limit to the size of clusters returned by the ``eom`` algorithm.
        Has no effect when using ``leaf`` clustering (where clusters are
        usually small regardless). Note that this should not be used if we
        want to predict the cluster labels for new points in future (e.g. using
        :func:`~flasc.prediction.approximate_predict`), as that function is
        not aware of this argument.

    label_sides_as_branches : bool, optional (default=False),
        When this flag is False, branches are only labelled for clusters with at
        least three branches (i.e., at least y-shapes). Clusters with only two
        branches represent l-shapes. The two branches describe the cluster's
        outsides growing towards each other. Enableing this flag separates these
        branches from each other in the produced labelling.

    Returns
    -------
    labels : np.ndarray, shape (n_samples, )
        Labels that differentiate all subgroups (clusters and branches). Noisy
        samples are given the label -1.

    probabilities : np.ndarray, shape (n_samples, )
        Probabilities considering both cluster and branch membership. Noisy
        samples are assigned 0.

    branch_labels : np.ndarray, shape (n_samples, )
        Branch labels for each point. Noisy samples are given the label -1.

    branch_probabilities : np.ndarray, shape (n_samples, )
        Branch membership strengths for each point. Noisy samples are
        assigned 0.

    branch_persistences : tuple (n_clusters)
        A branch persistence (eccentricity range) for each detected branch.

    cluster_approximation_graphs : tuple (n_clusters)
        The graphs used to detect branches in each cluster stored as a numpy
        array with four columns: source, target, centrality, mutual reachability
        distance. Points are labelled by their row-index into the input data.
        The edges contained in the graphs depend on the ``branch_detection_method``:
        - ``core``: Contains the edges that connect each point to all other
          points in a cluster within a mutual reachability distance lower than
          or equal to the point's core distance. This is an extension of the
          minimum spanning tree introducing only edges with equal distances. The
          reachability distance introduces ``num_points`` * ``min_samples`` of
          such edges.
        - ``full``: Contains all edges between points in each cluster with a
          mutual reachability distance lower than or equal to the distance of
          the most-distance point in each cluster. These graphs represent the
          0-dimensional simplicial complex of each cluster at the first point in
          the filtration where they contain all their points.

    cluster_condensed_trees : tuple (n_clusters)
        A condensed branch hierarchy for each cluster produced during the
        branch detection step. Data points are numbered with in-cluster ids.

    cluster_linkage_trees : tuple (n_clusters)
        A single linkage tree for each cluster produced during the branch
        detection step, in the scipy hierarchical clustering format.
        (see http://docs.scipy.org/doc/scipy/reference/cluster.hierarchy.html).
        Data points are numbered with in-cluster ids.

    cluster_centralities : np.ndarray, shape (n_samples, )
        Centrality values for each point in a cluster. Overemphasizes points'
        eccentricity within the cluster as the values are based on minimum
        spanning trees that do not contain the equally distanced edges resulting
        from the mutual reachability distance.

    cluster_points : list (n_clusters)
        The data point row indices for each cluster.

    References
    ----------
    .. [1] Bot, D. M., Peeters, J., Liesenborgs J., & Aerts, J. (2023, November).
       FLASC: A Flare-Sensitive Clustering Algorithm: Extending HDBSCAN* for
       Detecting Branches in Clusters. arXiv:2311.15887
    """
    # Check clusterer state
    if clusterer._min_spanning_tree is None:
        raise ValueError(
            "Clusterer does not have an explicit minimum spannning tree!"
            " Try fitting with branch_detection_data=True or"
            " gen_min_span_tree=True set."
        )
    if clusterer.branch_detection_data_ is None:
        raise ValueError(
            "Clusterer does not have branch detection data!"
            " Try fitting with branch_detection_data=True set,"
            " or run generate_branch_detection_data on the clusterer"
        )

    # Validate parameters
    if min_branch_size is None:
        min_branch_size = clusterer.min_cluster_size
    branch_selection_persistence = float(branch_selection_persistence)
    if not (np.issubdtype(type(min_branch_size), np.integer) and min_branch_size >= 2):
        raise ValueError(
            f"min_branch_size must be an integer greater or equal "
            f"to 2,  {min_branch_size} given."
        )
    if not (
        np.issubdtype(type(branch_selection_persistence), np.floating)
        and branch_selection_persistence >= 0.0
    ):
        raise ValueError(
            f"branch_selection_persistence must be a float greater or equal to "
            f"0.0, {branch_selection_persistence} given."
        )
    if branch_selection_method not in ("eom", "leaf"):
        raise ValueError(
            f"Invalid branch_selection_method: {branch_selection_method}\n"
            f'Should be one of: "eom", "leaf"\n'
        )
    if branch_detection_method not in ("core", "full"):
        raise ValueError(
            f"Invalid ``branch_detection_method``: {branch_detection_method}\n"
            'Should be one of: "core", "full"\n'
        )

    # Extract state
    memory = clusterer.memory
    if isinstance(memory, str):
        memory = Memory(memory, verbose=0)
    num_clusters = len(clusterer.cluster_persistence_)
    labels = clusterer.labels_
    probabilities = clusterer.probabilities_
    if not clusterer.branch_detection_data_.all_finite:
        finite_index = clusterer.branch_detection_data_.finite_index
        labels = labels[finite_index]
        probabilities = probabilities[finite_index]

    # Configure parallelization
    run_core = branch_detection_method == "core"
    num_jobs = clusterer.core_dist_n_jobs
    if num_jobs < 1:
        num_jobs = max(cpu_count() + 1 + num_jobs, 1)
    thread_pool = (
        SequentialPool() if run_core else Parallel(n_jobs=num_jobs, max_nbytes=None)
    )

    # Detect branches
    (
        cluster_points,
        cluster_centralities,
        cluster_linkage_trees,
        cluster_approximation_graphs,
    ) = memory.cache(_compute_branch_linkage, ignore=["thread_pool"])(
        labels,
        probabilities,
        clusterer._min_spanning_tree,
        clusterer.branch_detection_data_.tree,
        clusterer.branch_detection_data_.neighbors,
        clusterer.branch_detection_data_.core_distances,
        clusterer.branch_detection_data_.dist_metric,
        num_clusters,
        thread_pool,
        run_core=run_core,
    )
    (
        branch_labels,
        branch_probabilities,
        branch_persistences,
        cluster_condensed_trees,
    ) = memory.cache(_compute_branch_segmentation, ignore=["thread_pool"])(
        cluster_linkage_trees,
        thread_pool,
        min_branch_size=min_branch_size,
        allow_single_branch=allow_single_branch,
        branch_selection_method=branch_selection_method,
        branch_selection_persistence=branch_selection_persistence,
        max_branch_size=max_branch_size,
    )
    (
        labels,
        probabilities,
        branch_labels,
        branch_probabilities,
        cluster_centralities,
    ) = memory.cache(_update_labelling)(
        labels,
        probabilities,
        cluster_points,
        cluster_centralities,
        branch_labels,
        branch_probabilities,
        branch_persistences,
        label_sides_as_branches=label_sides_as_branches,
    )

    # Maintain data indices for non-finite data
    if not clusterer.branch_detection_data_.all_finite:
        internal_to_raw = clusterer.branch_detection_data_.internal_to_raw
        _remap_point_lists(cluster_points, internal_to_raw)
        _remap_edge_lists(cluster_approximation_graphs, internal_to_raw)

        num_points = len(clusterer.labels_)
        labels = _remap_labels(labels, finite_index, num_points)
        probabilities = _remap_probabilities(probabilities, finite_index, num_points)
        branch_labels = _remap_labels(branch_labels, finite_index, num_points)
        branch_probabilities = _remap_probabilities(
            branch_probabilities, finite_index, num_points
        )
        cluster_centralities = _remap_probabilities(
            cluster_centralities, finite_index, num_points
        )

    return (
        # Combined result
        labels,
        probabilities,
        # Branching result
        branch_labels,
        branch_probabilities,
        branch_persistences,
        # Clusters to branches
        cluster_approximation_graphs,
        cluster_condensed_trees,
        cluster_linkage_trees,
        cluster_centralities,
        cluster_points,
    )


def _compute_branch_linkage(
    cluster_labels,
    cluster_probabilities,
    min_spanning_tree,
    space_tree,
    neighbors,
    core_distances,
    dist_metric,
    num_clusters,
    thread_pool,
    run_core=False,
):
    result = thread_pool(
        delayed(_compute_branch_linkage_of_cluster)(
            cluster_labels,
            cluster_probabilities,
            min_spanning_tree,
            space_tree,
            neighbors,
            core_distances,
            dist_metric,
            run_core,
            cluster_id,
        )
        for cluster_id in range(num_clusters)
    )
    if len(result):
        return tuple(zip(*result))
    return (), (), (), ()


def _compute_branch_linkage_of_cluster(
    cluster_labels,
    cluster_probabilities,
    min_spanning_tree,
    space_tree,
    neighbors,
    core_distances,
    dist_metric,
    run_core,
    cluster_id,
):
    """Detect branches within one cluster."""
    # List points within cluster
    cluster_mask = cluster_labels == cluster_id
    cluster_points = np.where(cluster_mask)[0]
    in_cluster_ids = np.full(cluster_labels.shape[0], -1, dtype=np.double)
    in_cluster_ids[cluster_points] = np.arange(len(cluster_points), dtype=np.double)

    # Extract MST edges within cluster
    parent_mask = cluster_labels[min_spanning_tree[:, 0].astype(np.intp)] == cluster_id
    child_mask = cluster_labels[min_spanning_tree[:, 1].astype(np.intp)] == cluster_id
    cluster_mst = min_spanning_tree[parent_mask & child_mask]

    # Compute in cluster centrality
    points = space_tree.data.base[cluster_points]
    centroid = np.average(points, weights=cluster_probabilities[cluster_mask], axis=0)
    centralities = dist_metric.pairwise(centroid[None], points)[0, :]
    centralities = 1 / centralities

    # Construct cluster approximation graph
    if run_core:
        edges = _extract_core_cluster_graph(
            cluster_mst, core_distances, neighbors[cluster_points], in_cluster_ids
        )
    else:
        max_dist = cluster_mst.T[2].max()
        edges = _extract_full_cluster_graph(
            space_tree, core_distances, cluster_points, in_cluster_ids, max_dist
        )
    np.maximum(
        centralities[edges[:, 0].astype(np.intp)],
        centralities[edges[:, 1].astype(np.intp)],
        edges[:, 2],
    )

    # Extract centrality MST and compute single linkage
    centrality_mst = minimum_spanning_tree(
        coo_array(
            (edges[:, 2], (edges[:, 0].astype(np.int32), edges[:, 1].astype(np.int32))),
            shape=(len(cluster_points), len(cluster_points)),
        )
    ).tocoo()
    centrality_mst = np.column_stack(
        (centrality_mst.row, centrality_mst.col, centrality_mst.data)
    )
    centrality_mst = centrality_mst[np.argsort(centrality_mst.T[2]), :]
    linkage_tree = label(centrality_mst)

    # Re-label edges with data ids
    edges[:, 0] = cluster_points[edges[:, 0].astype(np.intp)]
    edges[:, 1] = cluster_points[edges[:, 1].astype(np.intp)]

    # Return values
    return cluster_points, centralities, linkage_tree, edges


def _extract_core_cluster_graph(
    cluster_spanning_tree,
    core_distances,
    neighbors,
    in_cluster_ids,
):
    """Create a graph connecting all points within each point's core distance."""
    # Allocate output (won't be filled completely)
    cluster_spanning_tree_view = cluster_spanning_tree
    num_points = neighbors.shape[0]
    num_neighbors = neighbors.shape[1]
    count = cluster_spanning_tree_view.shape[0]
    edges = np.zeros((count + num_points * num_neighbors, 4), dtype=np.double)

    # Fill (undirected) MST edges with within-cluster-ids
    mst_parents = in_cluster_ids[cluster_spanning_tree[:, 0].astype(np.intp)]
    mst_children = in_cluster_ids[cluster_spanning_tree[:, 1].astype(np.intp)]
    np.minimum(mst_parents, mst_children, edges[:count, 0])
    np.maximum(mst_parents, mst_children, edges[:count, 1])

    # Fill neighbors with within-cluster-ids
    core_parent = np.repeat(np.arange(num_points, dtype=np.double), num_neighbors)
    core_children = in_cluster_ids[neighbors.flatten()]
    np.minimum(core_parent, core_children, edges[count:, 0])
    np.maximum(core_parent, core_children, edges[count:, 1])

    # Fill mutual reachabilities
    edges[:count, 3] = cluster_spanning_tree[:, 2]
    np.maximum(
        core_distances[edges[count:, 0].astype(np.intp)],
        core_distances[edges[count:, 1].astype(np.intp)],
        edges[count:, 3],
    )

    # Extract unique edges that stay within the cluster
    edges = np.unique(edges[edges[:, 0] > -1.0, :], axis=0)
    return edges


def _extract_full_cluster_graph(
    space_tree, core_distances, cluster_points, in_cluster_ids, max_dist
):
    # Query KDTree/BallTree for neighours within the distance
    children_map, distances_map = space_tree.query_radius(
        space_tree.data.base[cluster_points], r=max_dist + 1e-8, return_distance=True
    )

    # Count number of returned edges per point
    num_children = np.zeros(len(cluster_points), dtype=np.intp)
    for i, children in enumerate(children_map):
        num_children[i] += len(children)

    # Create full edge list
    full_parents = np.repeat(
        np.arange(len(cluster_points), dtype=np.double), num_children
    )
    full_children = in_cluster_ids[np.concatenate(children_map)]
    full_distances = np.concatenate(distances_map)

    # Create output
    mask = (
        (full_children != -1.0)
        & (full_parents < full_children)
        & (full_distances <= max_dist)
    )
    edges = np.zeros((mask.sum(), 4), dtype=np.double)
    edges[:, 0] = full_parents[mask]
    edges[:, 1] = full_children[mask]
    np.maximum(
        np.maximum(
            core_distances[edges[:, 0].astype(np.intp)],
            core_distances[edges[:, 1].astype(np.intp)],
        ),
        full_distances[mask],
        edges[:, 3],
    )
    return edges


def _compute_branch_segmentation(
    cluster_linkage_trees,
    thread_pool,
    min_branch_size=5,
    allow_single_branch=False,
    branch_selection_method="eom",
    branch_selection_persistence=0.0,
    max_branch_size=0,
):
    """Extracts branches from the linkage hierarchies."""
    results = thread_pool(
        delayed(_compute_branch_segmentation_of_cluster)(
            cluster_linkage_tree,
            min_branch_size=min_branch_size,
            allow_single_branch=allow_single_branch,
            branch_selection_method=branch_selection_method,
            branch_selection_persistence=branch_selection_persistence,
            max_branch_size=max_branch_size,
        )
        for cluster_linkage_tree in cluster_linkage_trees
    )
    if len(results):
        return tuple(zip(*results))
    return (), (), (), ()


def _compute_branch_segmentation_of_cluster(
    cluster_linkage_tree,
    min_branch_size=5,
    allow_single_branch=False,
    branch_selection_method="eom",
    branch_selection_persistence=0.0,
    max_branch_size=0,
):
    """Select branches within one cluster."""
    condensed_tree = condense_tree(cluster_linkage_tree, min_branch_size)
    if branch_selection_persistence > 0.0:
        condensed_tree = simplify_branch_hierarchy(
            condensed_tree, branch_selection_persistence
        )
    stability = compute_stability(condensed_tree)
    (labels, probabilities, persistences) = get_branches(
        condensed_tree,
        stability,
        allow_single_branch=allow_single_branch,
        branch_selection_method=branch_selection_method,
        max_branch_size=max_branch_size,
    )
    # Reset noise labels to k-cluster
    labels[labels < 0] = len(persistences)
    return (labels, probabilities, persistences, condensed_tree)


def _update_labelling(
    cluster_labels,
    cluster_probabilities,
    cluster_points_,
    cluster_centralities_,
    branch_labels_,
    branch_probabilities_,
    branch_persistences_,
    label_sides_as_branches=False,
):
    """Updates the labelling with the detected branches."""
    # Allocate output
    num_points = len(cluster_labels)
    labels = -1 * np.ones(num_points, dtype=np.intp)
    probabilities = cluster_probabilities.copy()
    branch_labels = np.zeros(num_points, dtype=np.intp)
    branch_probabilities = np.ones(num_points, dtype=np.double)
    branch_centralities = np.zeros(num_points, dtype=np.double)

    # Compute the labels and probabilities
    running_id = 0
    for _points, _centralities, _labels, _probs, _pers in zip(
        cluster_points_,
        cluster_centralities_,
        branch_labels_,
        branch_probabilities_,
        branch_persistences_,
    ):
        num_branches = len(_pers)
        branch_centralities[_points] = _centralities
        if num_branches <= (1 if label_sides_as_branches else 2):
            labels[_points] = running_id
            running_id += 1
        else:
            labels[_points] = _labels + running_id
            branch_labels[_points] = _labels
            branch_probabilities[_points] = _probs
            probabilities[_points] += _probs
            probabilities[_points] /= 2
            running_id += num_branches + 1

    # Reorder other parts
    return (
        labels,
        probabilities,
        branch_labels,
        branch_probabilities,
        branch_centralities,
    )


def _remap_edge_lists(edge_lists, internal_to_raw):
    """
    Takes a list of edge lists and replaces the internal indices to raw indices.

    Parameters
    ----------
    edge_lists : list[np.ndarray]
        A list of numpy edgelists with the first two columns indicating
        datapoints.
    internal_to_raw: dict
        A mapping from internal integer index to the raw integer index.
    """
    for graph in edge_lists:
        for edge in graph:
            edge[0] = internal_to_raw[edge[0]]
            edge[1] = internal_to_raw[edge[1]]


def _remap_point_lists(point_lists, internal_to_raw):
    """
    Takes a list of points lists and replaces the internal indices to raw indices.

    Parameters
    ----------
    point_lists : list[np.ndarray]
        A list of numpy arrays with point indices.
    internal_to_raw: dict
        A mapping from internal integer index to the raw integer index.
    """
    for points in point_lists:
        for idx in range(len(points)):
            points[idx] = internal_to_raw[points[idx]]


def _remap_labels(old_labels, finite_index, num_points):
    """Creates new label array with infinite points set to -1."""
    new_labels = np.full(num_points, -1)
    new_labels[finite_index] = old_labels
    return new_labels


def _remap_probabilities(old_probs, finite_index, num_points):
    """Creates new probability array with infinite points set to 0."""
    new_probs = np.zeros(num_points)
    new_probs[finite_index] = old_probs
    return new_probs


class BranchDetector(BaseEstimator, ClusterMixin):
    """Performs a flare-detection post-processing step to detect branches within
    clusters [1]_.

    For each cluster, a graph is constructed connecting the data points based on
    their mutual reachability distances. Each edge is given a centrality value
    based on how far it lies from the cluster's center. Then, the edges are
    clustered as if that centrality was a distance, progressively removing the
    'center' of each cluster and seeing how many branches remain.

    Parameters
    ----------
    min_branch_size : int, optional (default=None)
        The minimum number of samples in a group for that group to be
        considered a branch; groupings smaller than this size will seen as
        points falling out of a branch. Defaults to the clusterer's min_cluster_size.

    allow_single_branch : bool, optional (default=False)
        Analogous to ``allow_single_cluster``.

    branch_detection_method : str, optional (default=``full``)
        Deteremines which graph is conctructed to detect branches with. Valid
        values are, ordered by increasing computation cost and decreasing
        sensitivity to noise:
        - ``core``: Contains the edges that connect each point to all other
          points within a mutual reachability distance lower than or equal to
          the point's core distance. This is the cluster's subgraph of the
          k-NN graph over the entire data set (with k = ``min_samples``).
        - ``full``: Contains all edges between points in each cluster with a
          mutual reachability distance lower than or equal to the distance of
          the most-distance point in each cluster. These graphs represent the
          0-dimensional simplicial complex of each cluster at the first point in
          the filtration where they contain all their points.

    branch_selection_method : str, optional (default='eom')
        The method used to select branches from the cluster's condensed tree.
        The standard approach for FLASC is to use the ``eom`` approach.
        Options are:
          * ``eom``
          * ``leaf``

    branch_selection_persistence: float, optional (default=0.0)
        An eccentricity persistence threshold. Branches with a persistence below
        this value will be merged. See [3]_ for more information. Note that this
        should not be used if we want to predict the cluster labels for new
        points in future (e.g. using approximate_predict), as the
        :func:`~flasc.prediction.approximate_predict` function is not aware of
        this argument.

    max_branch_size : int, optional (default=0)
        A limit to the size of clusters returned by the ``eom`` algorithm.
        Has no effect when using ``leaf`` clustering (where clusters are
        usually small regardless). Note that this should not be used if we
        want to predict the cluster labels for new points in future (e.g. using
        :func:`~flasc.prediction.approximate_predict`), as that function is
        not aware of this argument.

    label_sides_as_branches : bool, optional (default=False),
        When this flag is False, branches are only labelled for clusters with at
        least three branches (i.e., at least y-shapes). Clusters with only two
        branches represent l-shapes. The two branches describe the cluster's
        outsides growing towards each other. Enableing this flag separates these
        branches from each other in the produced labelling.

    Attributes
    ----------
    labels_ : np.ndarray, shape (n_samples, )
        Labels that differentiate all subgroups (clusters and branches). Noisy
        samples are given the label -1.

    probabilities_ : np.ndarray, shape (n_samples, )
        Probabilities considering both cluster and branch membership. Noisy
        samples are assigned 0.

    branch_labels_ : np.ndarray, shape (n_samples, )
        Branch labels for each point. Noisy samples are given the label -1.

    branch_probabilities_ : np.ndarray, shape (n_samples, )
        Branch membership strengths for each point. Noisy samples are
        assigned 0.

    branch_persistences_ : tuple (n_clusters)
        A branch persistence (eccentricity range) for each detected branch.

    cluster_approximation_graphs_ : tuple (n_clusters)
        The graphs used to detect branches in each cluster stored as a numpy
        array with four columns: source, target, centrality, mutual reachability
        distance. Points are labelled by their row-index into the input data.
        The edges contained in the graphs depend on the ``branch_detection_method``:
        - ``core``: Contains the edges that connect each point to all other
          points in a cluster within a mutual reachability distance lower than
          or equal to the point's core distance. This is an extension of the
          minimum spanning tree introducing only edges with equal distances. The
          reachability distance introduces ``num_points`` * ``min_samples`` of
          such edges.
        - ``full``: Contains all edges between points in each cluster with a
          mutual reachability distance lower than or equal to the distance of
          the most-distance point in each cluster. These graphs represent the
          0-dimensional simplicial complex of each cluster at the first point in
          the filtration where they contain all their points.

    cluster_condensed_trees_ : tuple (n_clusters)
        A condensed branch hierarchy for each cluster produced during the
        branch detection step. Data points are numbered with in-cluster ids.

    cluster_linkage_trees_ : tuple (n_clusters)
        A single linkage tree for each cluster produced during the branch
        detection step, in the scipy hierarchical clustering format.
        (see http://docs.scipy.org/doc/scipy/reference/cluster.hierarchy.html).
        Data points are numbered with in-cluster ids.

    cluster_centralities_ : np.ndarray, shape (n_samples, )
        Centrality values for each point in a cluster. Overemphasizes points'
        eccentricity within the cluster as the values are based on minimum
        spanning trees that do not contain the equally distanced edges resulting
        from the mutual reachability distance.

    cluster_points_ : list (n_clusters)
        The data point row indices for each cluster.

    References
    ----------
    .. [1] Bot, D. M., Peeters, J., Liesenborgs J., & Aerts, J. (2023, November).
       FLASC: A Flare-Sensitive Clustering Algorithm: Extending HDBSCAN* for
       Detecting Branches in Clusters. arXiv:2311.15887
    """

    def __init__(
        self,
        min_branch_size=None,
        allow_single_branch=False,
        branch_detection_method="full",
        branch_selection_method="eom",
        branch_selection_persistence=0.0,
        max_branch_size=0,
        label_sides_as_branches=False,
    ):
        self.min_branch_size = min_branch_size
        self.allow_single_branch = allow_single_branch
        self.branch_detection_method = branch_detection_method
        self.branch_selection_method = branch_selection_method
        self.branch_selection_persistence = branch_selection_persistence
        self.max_branch_size = max_branch_size
        self.label_sides_as_branches = label_sides_as_branches

        self._cluster_approximation_graphs = None
        self._cluster_condensed_trees = None
        self._cluster_linkage_trees = None
        self._branch_exemplars = None

    def fit(self, X, y=None):
        """
        Perform a flare-detection post-processing step to detect branches within
        clusters.

        Parameters
        ----------
        X : HDBSCAN
            A fitted HDBSCAN object with branch detection data generated.

        Returns
        -------
        self : object
            Returns self.
        """
        self._clusterer = X
        kwargs = self.get_params()
        (
            self.labels_,
            self.probabilities_,
            self.branch_labels_,
            self.branch_probabilities_,
            self.branch_persistences_,
            self._cluster_approximation_graphs,
            self._cluster_condensed_trees,
            self._cluster_linkage_trees,
            self.cluster_centralities_,
            self.cluster_points_,
        ) = detect_branches_in_clusters(X, **kwargs)

        return self

    def fit_predict(self, X, y=None):
        """
        Perform a flare-detection post-processing step to detect branches within
        clusters [1]_.

        Parameters
        ----------
        X : HDBSCAN
            A fitted HDBSCAN object with branch detection data generated.

        Returns
        -------
        labels : ndarray, shape (n_samples, )
            subgroup labels differentiated by cluster and branch.
        """
        self.fit(X, y)
        return self.labels_

    def weighted_centroid(self, label_id, data=None):
        """Provides an approximate representative point for a given branch.
        Note that this technique assumes a euclidean metric for speed of
        computation. For more general metrics use the ``weighted_medoid`` method
        which is slower, but can work with the metric the model trained with.

        Parameters
        ----------
        label_id: int
            The id of the cluster to compute a centroid for.

        data : np.ndarray (n_samples, n_features), optional (default=None)
            A dataset to use instead of the raw data that was clustered on.

        Returns
        -------
        centroid: array of shape (n_features,)
            A representative centroid for cluster ``label_id``.
        """
        if self.labels_ is None:
            raise AttributeError("Model has not been fit to data")
        if self._clusterer._raw_data is None and data is None:
            raise AttributeError("Raw data not available")
        if label_id == -1:
            raise ValueError(
                "Cannot calculate weighted centroid for -1 cluster "
                "since it is a noise cluster"
            )
        if data is None:
            data = self._clusterer._raw_data
        mask = self.labels_ == label_id
        cluster_data = data[mask]
        cluster_membership_strengths = self.probabilities_[mask]

        return np.average(cluster_data, weights=cluster_membership_strengths, axis=0)

    def weighted_medoid(self, label_id, data=None):
        """Provides an approximate representative point for a given branch.

        Note that this technique can be very slow and memory intensive for large
        clusters. For faster results use the ``weighted_centroid`` method which
        is faster, but assumes a euclidean metric.

        Parameters
        ----------
        label_id: int
            The id of the cluster to compute a medoid for.

        data : np.ndarray (n_samples, n_features), optional (default=None)
            A dataset to use instead of the raw data that was clustered on.

        Returns
        -------
        centroid: array of shape (n_features,)
            A representative medoid for cluster ``label_id``.
        """
        if self.labels_ is None:
            raise AttributeError("Model has not been fit to data")
        if self._clusterer._raw_data is None and data is None:
            raise AttributeError("Raw data not available")
        if label_id == -1:
            raise ValueError(
                "Cannot calculate weighted centroid for -1 cluster "
                "since it is a noise cluster"
            )
        if data is None:
            data = self._clusterer._raw_data
        mask = self.labels_ == label_id
        cluster_data = data[mask]
        cluster_membership_strengths = self.probabilities_[mask]

        dist_metric = self._clusterer.branch_detection_data_.dist_metric
        dist_mat = dist_metric.pairwise(cluster_data) * cluster_membership_strengths
        medoid_index = np.argmin(dist_mat.sum(axis=1))
        return cluster_data[medoid_index]

    @property
    def cluster_approximation_graph_(self):
        """See :class:`~hdbscan.branches.BranchDetector` for documentation."""
        if self._cluster_approximation_graphs is None:
            raise AttributeError(
                "No cluster approximation graph was generated; try running fit first."
            )
        return ApproximationGraph(
            self._cluster_approximation_graphs,
            self.labels_,
            self.probabilities_,
            self._clusterer.labels_,
            self._clusterer.probabilities_,
            self.cluster_centralities_,
            self.branch_labels_,
            self.branch_probabilities_,
            self._clusterer._raw_data,
        )

    @property
    def cluster_condensed_trees_(self):
        """See :class:`~hdbscan.branches.BranchDetector` for documentation."""
        if self._cluster_condensed_trees is None:
            raise AttributeError(
                "No cluster condensed trees were generated; try running fit first."
            )
        return [
            CondensedTree(tree, self.branch_selection_method, self.allow_single_branch)
            for tree in self._cluster_condensed_trees
        ]

    @property
    def cluster_linkage_trees_(self):
        """See :class:`~hdbscan.branches.BranchDetector` for documentation."""
        if self._cluster_linkage_trees is None:
            raise AttributeError(
                "No cluster linkage trees were generated; try running fit first."
            )
        return [SingleLinkageTree(tree) for tree in self._cluster_linkage_trees]

    @property
    def branch_exemplars_(self):
        """See :class:`~hdbscan.branches.BranchDetector` for documentation."""
        if self._branch_exemplars is not None:
            return self._branch_exemplars
        if self._clusterer._raw_data is None:
            raise AttributeError(
                "Branch exemplars not available with precomputed " "distances."
            )
        if self._cluster_condensed_trees is None:
            raise AttributeError("No branches detected; try running fit first.")

        num_clusters = len(self._cluster_condensed_trees)
        branch_cluster_trees = [
            branch_tree[branch_tree["child_size"] > 1]
            for branch_tree in self._cluster_condensed_trees
        ]
        selected_branch_ids = [
            sorted(branch_tree._select_clusters())
            for branch_tree in self.cluster_condensed_trees_
        ]

        self._branch_exemplars = [None] * num_clusters

        for i, points in enumerate(self.cluster_points_):
            selected_branches = selected_branch_ids[i]
            if len(selected_branches) <= (1 if self.label_sides_as_branches else 2):
                continue

            self._branch_exemplars[i] = []
            raw_condensed_tree = self._cluster_condensed_trees[i]

            for branch in selected_branches:
                _branch_exemplars = np.array([], dtype=np.intp)
                for leaf in recurse_leaf_dfs(branch_cluster_trees[i], np.intp(branch)):
                    leaf_max_lambda = raw_condensed_tree["lambda_val"][
                        raw_condensed_tree["parent"] == leaf
                    ].max()
                    candidates = raw_condensed_tree["child"][
                        (raw_condensed_tree["parent"] == leaf)
                        & (raw_condensed_tree["lambda_val"] == leaf_max_lambda)
                    ]
                    _branch_exemplars = np.hstack([_branch_exemplars, candidates])
                ids = points[_branch_exemplars]
                self._branch_exemplars[i].append(self._clusterer._raw_data[ids, :])

        return self._branch_exemplars


def approximate_predict_branch(branch_detector, points_to_predict):
    """Predict the cluster and branch label of new points.

    Extends ``approximate_predict`` to also predict in which branch
    new points lie (if the cluster they are part of has branches).

    Parameters
    ----------
    branch_detector : BranchDetector
        A clustering object that has been fit to vector inpt data.

    points_to_predict : array, or array-like (n_samples, n_features)
        The new data points to predict cluster labels for. They should
        have the same dimensionality as the original dataset over which
        clusterer was fit.

    Returns
    -------
    labels : array (n_samples,)
        The predicted cluster and branch labels.

    probabilities : array (n_samples,)
        The soft cluster scores for each.

    cluster_labels : array (n_samples,)
        The predicted cluster labels.

    cluster_probabilities : array (n_samples,)
        The soft cluster scores for each.

    branch_labels : array (n_samples,)
        The predicted cluster labels.

    branch_probabilities : array (n_samples,)
        The soft cluster scores for each.
    """

    cluster_labels, cluster_probabilities, connecting_points = approximate_predict(
        branch_detector._clusterer, points_to_predict, return_connecting_points=True
    )

    num_predict = len(points_to_predict)
    labels = np.empty(num_predict, dtype=np.intp)
    probabilities = np.zeros(num_predict, dtype=np.double)
    branch_labels = np.zeros(num_predict, dtype=np.intp)
    branch_probabilities = np.ones(num_predict, dtype=np.double)

    min_num_branches = 2 if not branch_detector.label_sides_as_branches else 1
    for i, (label, prob, connecting_point) in enumerate(
        zip(cluster_labels, cluster_probabilities, connecting_points)
    ):
        if label < 0:
            labels[i] = -1
        elif len(branch_detector.branch_persistences_[label]) <= min_num_branches:
            labels[i] = label
            probabilities[i] = prob
        else:
            labels[i] = branch_detector.labels_[connecting_point]
            branch_labels[i] = branch_detector.branch_labels_[connecting_point]
            branch_probabilities[i] = branch_detector.branch_probabilities_[
                connecting_point
            ]
            probabilities[i] = (prob + branch_probabilities[i]) / 2
    return (
        labels,
        probabilities,
        cluster_labels,
        cluster_probabilities,
        branch_labels,
        branch_probabilities,
    )


class SequentialPool:
    """API of a Joblib Parallel pool but sequential execution"""

    def __init__(self):
        self.n_jobs = 1

    def __call__(self, jobs):
        return [fun(*args, **kwargs) for (fun, args, kwargs) in jobs]



import json
import re
import numpy as np
from typing import Optional, List, Dict, Any

class XYZ:
    def __init__(self, file_location: str = None, name: str = None, **kwargs):
        """
        Initializes the XYZ object.

        Parameters:
            file_location (str): The default path for reading/writing XYZ files.
            name (str): An identifier for this XYZ object.
            **kwargs: Additional keyword arguments passed to the parent constructor.
        """
        self.property_info = None
        self.metadata = None  # Initialize metadata; can hold a dict of lists or any structured data.

    def export_as_xyz(
        self,
        file_location: str = None,
        save_to_file: str = "w",
        verbose: bool = False,
        species: bool = True,
        position: bool = True,
        energy: bool = True,
        forces: bool = True,
        charge: bool = True,
        magnetization: bool = True,
        lattice: bool = True,
        pbc: bool = True,
        time: bool = True,
        fixed: bool = True,
        class_ID: bool = True,
        position_tag: str = "pos",
        forces_tag: str = "forces",
        charge_tag: str = "charge",
        magnetization_tag: str = "magnetization",
        energy_tag: str = "energy",
        fixed_tag: str = "fixed",
        time_tag: str = "time",
        pbc_tag: str = "pbc",
        classID_tag: str = "class_ID",
    ) -> str:
        out = file_location or self.file_location

        # 1) group elements if needed
        #if hasattr(self, "group_elements_and_positions"):
        #    self.group_elements_and_positions()
        # Convenience: choose constraints source if present
        ac = getattr(self, "atomicConstraints", None)
        if ac is None:
            ac = getattr(self, fixed_tag, None)  # allow self.fixed as source too

        # 2) precompute booleans
        inc = {
            "lattice":   lattice   and getattr(self, "latticeVectors", None) is not None,
            "species":   species   and getattr(self, "atomLabelsList", None) is not None,
            "pos":       position  and getattr(self, "atomPositions", None) is not None,
            "forces":    forces    and getattr(self, "total_force", None) is not None,
            "charge":    charge    and getattr(self, "charge", None) is not None,
            "magnet":    magnetization and getattr(self, "magnetization", None) is not None,
            "energy":    energy    and getattr(self, "E", None) is not None,
            "pbc":       pbc       and getattr(self, "latticeVectors", None) is not None,
            "time":      time      and getattr(self, "time", None) is not None,
            "fixed":     fixed     and getattr(self, "selectiveDynamics", False),
            "classID":   class_ID  and getattr(self, "class_ID", None) is not None,
            "meta":      getattr(self, "metadata", None) is not None
        }

        # Determine number of fixed columns (1 for (N,), 3 for (N,3))
        n_fixed_cols = 0
        if inc["fixed"]:
            ac_arr = np.asarray(ac)
            if ac_arr.ndim == 1:
                n_fixed_cols = 1
            elif ac_arr.ndim == 2 and ac_arr.shape[1] in (1, 3):
                n_fixed_cols = int(ac_arr.shape[1])
            else:
                # Fallback: flatten one row to decide; keep it general
                n_fixed_cols = int(np.asarray(ac_arr[0]).size)
        
        # 3) build header
        hdr = []
        if inc["lattice"]:
            vals = " ".join(map(str, self.latticeVectors.flatten()))
            hdr.append(f'Lattice="{vals}"')

        props = []
        if inc["species"]: props.append("species:S:1")
        if inc["pos"]:      props.append(f"{position_tag}:R:3")
        if inc["forces"]:   props.append(f"{forces_tag}:R:3")
        if inc["charge"]:   props.append(f"{charge_tag}:R:1")
        if inc["magnet"]:   props.append(f"{magnetization_tag}:R:1")
        if inc["fixed"]:    props.append(f"{fixed_tag}:I:{n_fixed_cols}")  
        if inc["classID"]:  props.append(f"{classID_tag}:I:1")
        if props:
            hdr.append("Properties=" + ":".join(props))
        if inc["energy"]:
            hdr.append(f"{energy_tag}={self.E}")
        if inc["pbc"]:
            hdr.append(f'{pbc_tag}="T T T"')
        if inc["time"]:
            hdr.append(f"{time_tag}={self.time}")
        if inc["meta"]:
            hdr.append(" ".join(
                f"{k}={json.dumps(v, separators=(',', ':'), ensure_ascii=False)}"
                for k, v in self.metadata.items()
            ))
        header_line = " ".join(hdr)

        # 4) build atom lines
        lines = []

        for i in range(self.atomCount):
            cols = []
            if inc["species"]:
                #print( self.atomLabelsList, i, self.atomCount )
                cols.append(self.atomLabelsList[i])
            if inc["pos"]:
                x, y, z = self.atomPositions[i]
                cols += [f"{c:.10f}" for c in (x, y, z)]
            if inc["forces"]:
                cols += [f"{f:.10f}" for f in self.total_force[i]]
            if inc["charge"]:
                ch = self.charge[i] if np.ndim(self.charge) == 1 else self.charge[i, -1]
                cols.append(f"{ch:.10f}")
            if inc["magnet"]:
                mg = self.magnetization[i] if np.ndim(self.magnetization) == 1 else self.magnetization[i, -1]
                cols.append(f"{mg:.10f}")
            #if inc["fixed"]:
            #    cols += str(int(self.atomicConstraints[i])) 
            if inc["fixed"]:
                row = np.asarray(ac[i])
                if row.shape == ():  # numpy scalar
                    cols.append(str(int(bool(row.item()))))
                else:
                    row = row.reshape(-1)
                    # Normalise any bool/float to 0/1 ints
                    cols += [str(int(bool(x))) for x in row[:n_fixed_cols]]

            if inc["classID"]:
                cid = self.class_ID[i] if i < len(self.class_ID) else -1
                cols.append(str(cid))
            lines.append(" ".join(cols))

        # 5) assemble
        content = f"{self.atomCount}\n{header_line}\n" + "\n".join(lines) + "\n"

        # 6) write if needed
        if out and save_to_file != "none":
            with open(out, save_to_file) as f:
                f.write(content)
            if verbose:
                print(f"XYZ content saved to {out}")
        return content

    def read_XYZ(self,
                 file_location: Optional[str] = None,
                 lines: Optional[List[str]] = None,
                 verbose: bool = False,
                 tags: Optional[Dict[str, str]] = None,
                 **kwargs) -> Dict[str, Any]:
        path = file_location or self.file_location
        if not path:
            raise ValueError("Must provide a file_location")

        # Read raw lines if caller passed them, else open file
        if lines is None:
            with open(path, 'r') as f:
                # 1) atom count
                atomCount = int(f.readline().strip())
                # 2) header
                header = f.readline().strip()
                self._parse_header(header)
                # 3) now f yields exactly the atomCount lines (or more; we stop at atomCount)
                raw_lines = (next(f) for _ in range(atomCount))
        else:
            atomCount = int(lines[0].strip())
            self._parse_header(lines[1].strip())
            raw_lines = iter(lines[2:2+atomCount])

        # Build storage lists
        if self.property_info is None:
            # Fallback for standard XYZ: species, x, y, z
            self.property_info = [("species", "S", 1), ("pos", "R", 3)]
        
        props = self.property_info  # list of (name, dtype, ncols)
        storage: Dict[str, List[Any]] = { name: [] for name,_,_ in props }

        # Single‐pass parse
        for raw in raw_lines:
            tok = raw.split()
            idx = 0
            for name, dtype, ncols in props:
                chunk = tok[idx:idx+ncols]
                idx += ncols
                if dtype == 'S':
                    # string property
                    storage[name].append(chunk[0] if ncols == 1 else chunk)
                else:
                    # numeric property
                    nums = list(map(float, chunk))
                    storage[name].append(nums[0] if ncols == 1 else nums)

        # Bulk convert to numpy arrays
        for name, dtype, ncols in props:
            data = storage[name]
            if dtype == 'S':
                arr = np.array(data, dtype=object)
            elif dtype == 'R':
                arr = np.array(data, dtype=float).reshape(atomCount, ncols)
            else:  # 'I'
                arr = np.array(data, dtype=int).reshape(atomCount, ncols)
            setattr(self, name, arr)

        # Aliases for downstream code
        self.atomLabelsList = getattr(self, 'species', None)
        self.atomPositions  = getattr(self, 'pos', None)
        self.total_force     = getattr(self, 'forces', None)
        self.charge          = getattr(self, 'charge', None)
        self.magnetization   = getattr(self, 'magnetization', None)
        self.atomCount       = atomCount

        return {
            'position':      self.atomPositions,
            'atomCount':     self.atomCount,
            'species':       self.atomLabelsList,
            'forces':        self.total_force,
            'charge':        self.charge,
            'magnetization': self.magnetization,
            'latticeVectors': getattr(self, 'latticeVectors', None),
            'energy':         getattr(self, 'E', None),
            'pbc':            getattr(self, 'pbc', None),
            'metadata':       getattr(self, 'metadata', None)
        }

    def _parse_header(self, header_line: str) -> None:
        """
        Parses the header line of the extended XYZ format, extracting information such as
        Lattice, Properties, energy, pbc, and metadata (if present).
        """
        # Use a regex to find key=value pairs. Values may be in quotes or unquoted.
        header_parts = re.findall(r'(\w+)=("[^"]+"|[^\s]+)', header_line)

        for key, value in header_parts:
            if key == 'Lattice':
                # Convert the flattened numeric string into a 3x3 numpy array
                self.latticeVectors = np.array(list(map(float, value.strip('"').split()))).reshape(3, 3)
            elif key == 'Properties':
                # This sets self.property_info so we know how many columns each property occupies
                self._parse_properties(value.strip('"'))
            elif key == 'energy':
                self.E = float(value)
            elif key == 'pbc':
                # Expecting strings like "T T T" to indicate True/False for each dimension
                self.pbc = [v.lower() == 't' for v in value.strip('"').split()]
            elif key == 'info_system':
                # Attempt to parse JSON back into a Python object
                self.info_system = json.loads(value.strip('"'))
            else:
                # If we have other custom keys, store them in an info dictionary
                if not hasattr(self, 'metadata'):
                    self.metadata = {}
                if not isinstance(self.metadata, dict):
                    self.metadata = {}
                self.metadata[key] = self._coerce_header_value(value)

    def _parse_properties(self, properties_str: str) -> None:
        """
        Parses the 'Properties=' definition to understand the data columns that follow
        in the atomic section.

        E.g., a string like: species:S:1:pos:R:3:forces:R:3
        tells us that the columns are:

            1) species (string)  # 1 column
            2) pos (real)        # 3 columns
            3) forces (real)     # 3 columns
        """
        self.property_info = []
        parts = properties_str.split(':')
        for i in range(0, len(parts), 3):
            name, dtype, ncols = parts[i:i+3]
            self.property_info.append((name, dtype, int(ncols)))

    def _parse_atom_data(self, atom_lines: List[str]) -> None:
        """
        Parses the per-atom data lines according to the structure defined
        in self.property_info from the header line.

        Parameters:
            atom_lines (List[str]): The lines following the header in the extended XYZ file.
        """
        # Split each line by whitespace and form a 2D array
        clean_lines = [line.strip() for line in atom_lines if line.strip()]
        data = np.array([line.split() for line in clean_lines])

        if len(data) != self.atomCount:
            raise ValueError(f"Number of data lines ({len(data)}) does not match atomCount ({self.atomCount})")

        col_index = 0
        for name, dtype, ncols in self.property_info:
            end_index = col_index + ncols
            if end_index > data.shape[1]:
                raise ValueError(f"Not enough columns in data for property '{name}'")

            if dtype == 'S':
                # String columns
                setattr(self, name, data[:, col_index])
            elif dtype == 'R':
                # Float columns
                setattr(self, name, data[:, col_index:end_index].astype(float))
            elif dtype == 'I':
                # Integer columns
                if ncols == 1:
                    setattr(self, name, data[:, col_index].astype(int))
                else:
                    setattr(self, name, data[:, col_index:end_index].astype(int))
            col_index = end_index

        # Assign aliases to known property names
        self.atomLabelsList = getattr(self, 'species', None)
        self.atomPositions = getattr(self, 'pos', None)
        self.total_force = getattr(self, 'forces', None)
        self.charge = getattr(self, 'charge', None)
        self.magnetization = getattr(self, 'magnetization', None)

    @staticmethod
    def read_file(file_location: str, strip: bool = True) -> List[str]:
        """
        Reads the content of a file and returns its lines.

        Parameters:
            file_location (str): The location of the file to read.
            strip (bool): If True, strips whitespace from each line.

        Returns:
            List[str]: All lines from the file.
        """
        with open(file_location, 'r') as f:
            if strip:
                return [line.strip() for line in f]
            else:
                return [line for line in f]

    @staticmethod
    def _coerce_header_value(v: str):
        """
        Convert a header value token to an appropriate Python type.
        Handles quoted/unquoted values, JSON scalars/arrays/objects,
        T/F booleans, ints, and floats; otherwise returns a string.
        """
        # 1) remove one layer of double quotes, if present
        quoted = len(v) >= 2 and v[0] == '"' and v[-1] == '"'
        inner = v[1:-1] if quoted else v

        # 2) try JSON first (handles numbers, lists like [1], dicts, true/false/null)
        try:
            return json.loads(inner)
        except Exception:
            pass

        # 3) try T/F booleans (common in materials headers)
        if inner in ("T", "F", "t", "f"):
            return inner.lower().startswith("t")

        # 4) try integer
        if re.fullmatch(r"[+-]?\d+", inner):
            try:
                return int(inner)
            except Exception:
                pass

        # 5) try float (incl. scientific notation)
        if re.fullmatch(r"[+-]?(?:\d+\.\d*|\.\d+|\d+)(?:[eE][+-]?\d+)?", inner):
            try:
                return float(inner)
            except Exception:
                pass

        # 6) default: plain string
        return inner



class SI:
    def __init__(self, file_location:str=None, name:str=None, **kwargs):
        pass
        
    def read_SI(self, file_location:str=None):
        try:
            import numpy as np
        except ImportError as e:
            import sys
            sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
            del sys

        # read files commondly presente in the SI
        file_location = file_location if type(file_location) == str else self._file_location

        lines = [n for n in self.read_file() ]
                    
        # Flags to indicate which section of the file we are in
        reading_lattice_vectors = False
        reading_atomic_positions = False

        self._latticeVectors = []
        self._atomLabelsList = []
        self._atomPositions = []

        for line in lines:
            # Remove leading and trailing whitespaces
            line = line.strip()

            # Check for section headers
            if "Supercell lattice vectors" in line:
                reading_lattice_vectors = True
                reading_atomic_positions = False
                continue
            elif "Atomic positions" in line:
                reading_lattice_vectors = False
                reading_atomic_positions = True
                continue
            
            # Read data based on current section
            if reading_lattice_vectors:
                vector = [float(x) for x in line.split(",")]
                self._latticeVectors.append(vector)
            elif reading_atomic_positions:
                elements = line.split()
                self._atomLabelsList.append(elements[0])
                self._atomPositions.append([ float(n) for n in elements[1:] ])

        self._atomPositions = np.array(self._atomPositions)             
        self._atomLabelsList = np.array(self._atomLabelsList)             
        self._latticeVectors = np.array(self._latticeVectors)             
        self._atomCoordinateType = 'Cartesian'
        self._atomicConstraints = np.ones_like(self._atomPositions)
        self._atomCount = self._atomPositions.shape[0]
        self._selectiveDynamics = True
        self._scaleFactor = [1]

        return True



"""
Chemical Feature Relevance Analysis with SAGE
=============================================

This script provides an abstract, complete example of how to perform 
feature analysis on chemical data using the SAGE library.

It demonstrates a standard workflow:
1. Generating/Loading abstract chemical features (element fractions, bond statistics).
2. Cleaning features (removing NaNs, constant columns).
3. Analyzing feature redundancy (Correlation Clustering).
4. Identifying relevant features using Stability Selection (Robust XGBoost).
5. Visualizing/Reporting the critical descriptors.

This example uses synthetic data to simulate a scenario where we want to identify 
which chemical representations (e.g., fractional compositions vs. bond lengths) 
drive a target property like Formation Energy.
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from xgboost import XGBRegressor
from sklearn.tree import DecisionTreeRegressor, export_text
from sklearn.inspection import PartialDependenceDisplay
import shap
import os

# SAGE Library Imports
try:
    from sage_lib.partition.partition_analysis.featurizer.analysis.feature_analysis import (
        clean_features,
        cluster_features,
        pick_representatives,
        stability_selection
    )
    from sage_lib.partition.partition_analysis.featurizer.analysis.xgb_postanalysis import (
        xgb_importances
    )
except ImportError:
    # If running standalone without the full package structure, we might need path adjustments
    import sys
    import os
    sys.path.append(os.path.abspath("../../../../..")) # Adjust based on actual location
    from sage_lib.partition.partition_analysis.featurizer.analysis.feature_analysis import (
        clean_features,
        cluster_features,
        pick_representatives,
        stability_selection
    )
    from sage_lib.partition.partition_analysis.featurizer.analysis.xgb_postanalysis import (
        xgb_importances
    )

def generate_abstract_chemical_data(n_samples=500, random_state=42):
    """
    Generates a synthetic dataset representing chemical features.
    
    The target property 'Formation_Energy' is constructed to depend on 
    specific interactions with optimized points and penalties.
    
    Rules implemented:
    1. Optimal Point: C-H bond length is optimal (lowest energy) at 1.10 Å. Devation increases energy (parabola).
    2. Destabilizing Factor: Oxygen fraction > 0.6 drastically increases energy (instability).
    3. Linear Benefit: Higher Hydrogen fraction linearly decreases energy (stabilizing).
    """
    np.random.seed(random_state)
    
    # 1. Fundamental Chemical Features (Independent Variables)
    frac_C = np.random.uniform(0.1, 0.8, n_samples)
    frac_H = np.random.uniform(0.1, 0.8, n_samples)
    scale = (frac_C + frac_H + np.random.uniform(0.0, 0.2, n_samples))
    frac_C /= scale
    frac_H /= scale
    frac_O = 1.0 - (frac_C + frac_H) # Balance
    
    # 2. Structural Features (Bond lengths in Angstroms)
    # C-H bond length correlates slightly with C fraction but has variance
    mean_dist_CH = 1.09 + (frac_C * 0.1) + np.random.normal(0, 0.05, n_samples)
    mean_dist_OH = 0.96 + np.random.normal(0, 0.05, n_samples)
    
    # 3. Geometric Features
    vol_per_atom = 10.0 + (frac_C * 5.0) + np.random.normal(0, 1.0, n_samples)
    
    # 4. Redundant/Correlated Features
    carbon_content_pct = frac_C * 100.0
    
    # 5. Noise Features
    noise_1 = np.random.rand(n_samples)
    noise_2 = np.random.rand(n_samples)
    
    # 6. Construct Target: Formation Energy (eV/atom)
    formation_energy = np.zeros(n_samples)
    
    # Rule A: Parabolic stability for C-H bond (Optimal at 1.10)
    # E += 20 * (x - 1.10)^2
    formation_energy += 20.0 * (mean_dist_CH - 1.10)**2
    
    # Rule B: Linear stabilization from Hydrogen
    formation_energy -= 2.0 * frac_H
    
    # Rule C: Instability if Oxygen > 0.6 (Step function / penalty)
    formation_energy += np.where(frac_O > 0.6, 5.0, 0.0)
    
    # Add noise
    formation_energy += np.random.normal(0, 0.1, n_samples)
    
    # Create DataFrame
    data = {
        'frac_C': frac_C,
        'frac_H': frac_H,
        'frac_O': frac_O,
        'mean_dist_C-H': mean_dist_CH,
        'mean_dist_O-H': mean_dist_OH,
        'vol_per_atom': vol_per_atom,
        'pct_C': carbon_content_pct, # Redundant
        'noise_metric_1': noise_1,   # Irrelevant
        'noise_metric_2': noise_2    # Irrelevant
    }
    
    X = pd.DataFrame(data)
    y = pd.Series(formation_energy, name="Formation_Energy")
    
    return X, y

def analyze_chemical_relevance(X, y, output_dir="analysis_results"):
    """
    Performs the standard SAGE feature analysis workflow with advanced interpretation.
    """
    os.makedirs(output_dir, exist_ok=True)
    
    print("=========================================================")
    print(" SAGE Chemical Feature Analysis & Rule Extraction")
    print("=========================================================")
    
    # 1. Clean & Cluster
    X_clean = clean_features(X, nan_strategy='median', drop_thresh=0.2)
    clusters = cluster_features(X_clean, corr_threshold=0.95)
    representatives = pick_representatives(X_clean, clusters)
    X_reduced = X_clean[representatives]
    print(f" -> Selected Representative Features: {representatives}")

    # 2. Stability Selection
    print("\n[Step 2] Stability Selection (Identifying Robust Drivers)...")
    stability_scores = stability_selection(X_reduced, y.values, n_runs=20)
    relevant_features = stability_scores[stability_scores > 0.4].index.tolist()
    print(f" -> Critical Features identified: {relevant_features}")
    
    # 3. Model Training (for Explanation)
    print("\n[Step 3] Fitting Interpretability Model (XGBoost)...")
    model = XGBRegressor(n_estimators=200, max_depth=4, learning_rate=0.05)
    model.fit(X_reduced[relevant_features], y)
    
    # =========================================================
    # ADVANCED ANALYSIS: Finding Optimal Points & Rules
    # =========================================================
    
    # A. SHAP Analysis (Directionality & Contribution)
    print("\n[Step 4] SHAP Analysis (Stability Contributors)...")
    explainer = shap.TreeExplainer(model)
    shap_values = explainer.shap_values(X_reduced[relevant_features])
    
    # Summary Plot
    plt.figure(figsize=(10, 6))
    shap.summary_plot(shap_values, X_reduced[relevant_features], show=False)
    plt.title("SHAP Summary: Feature Impact on Formation Energy")
    plt.tight_layout()
    plt.savefig(f"{output_dir}/shap_summary.png")
    plt.close()
    print(f" -> Saved SHAP summary to {output_dir}/shap_summary.png")
    
    # B. Partial Dependence Plots (PDP) (Finding Optimal Values)
    print("\n[Step 5] Partial Dependence Plots (Finding Optimal Values)...")
    # We plot PDP for the top 3 relevant features
    top_feats = relevant_features[:3] 
    
    fig, ax = plt.subplots(figsize=(12, 4))
    display = PartialDependenceDisplay.from_estimator(
        model, 
        X_reduced[relevant_features], 
        top_feats, 
        kind="average", 
        ax=ax
    )
    plt.suptitle("Partial Dependence: Identifying Optimal Operating Points", y=1.05)
    plt.tight_layout()
    plt.savefig(f"{output_dir}/pdp_optimal_points.png")
    plt.close()
    print(f" -> Saved PDP plots to {output_dir}/pdp_optimal_points.png")
    
    # C. Rule Extraction (Surrogate Decision Tree)
    print("\n[Step 6] Extracting Fundamental Rules (Decision Tree Surrogate)...")
    # We fit a simple Decision Tree to mimic the complex XGBoost model
    # This gives us explicit "If-Then" rules.
    tree = DecisionTreeRegressor(max_depth=3)
    tree.fit(X_reduced[relevant_features], model.predict(X_reduced[relevant_features]))
    
    rules = export_text(tree, feature_names=relevant_features)
    print("\n--- EXTRACTED RULES (Approximation) ---")
    print(rules)
    
    with open(f"{output_dir}/derived_rules.txt", "w") as f:
        f.write(rules)

def main():
    # 1. Generate Data with Hidden Rules
    X, y = generate_abstract_chemical_data(n_samples=500)
    
    # 2. Run Analysis
    analyze_chemical_relevance(X, y)
    
    print("\nAnalysis Complete.")
    print("Check 'analysis_results/' for plots and rule text.")

if __name__ == "__main__":
    main()



# ============================
# memory.py — In-memory (list/dict) backend
# ============================
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

from .backend import StorageBackend, StorageType

# -----------------------------
# In-memory (list/dict) backend
# -----------------------------
class MemoryStorage(StorageBackend):
    """
    Generic in-memory storage.

    Notes
    -----
    - The container can be either a list (sequential storage) or a dict mapping
      integer IDs to objects.
    - IDs are always integers, automatically assigned if using a list.
    """

    def __init__(self, initial: StorageType | None = None) -> None:
        self._data: StorageType = initial if initial is not None else []
        self._meta: Dict[int, Dict[str, Any]] = {}

    # ---------------------------------------
    # Core CRUD interface
    # ---------------------------------------
    def add(self, obj: Any, metadata: Optional[Dict[str, Any]] = None) -> int:
        if isinstance(self._data, list):
            self._data.append(obj)
            idx = len(self._data) - 1
        else:
            idx = max(self._data.keys(), default=-1) + 1
            self._data[idx] = obj
        if metadata is not None:
            self._meta[idx] = metadata
        return idx

    def add_many(self, objs: List[Any], metadata: Optional[List[Dict[str, Any]]] = None) -> List[int]:
        if not objs:
            return []
            
        if isinstance(self._data, list):
            start_idx = len(self._data)
            self._data.extend(objs)
            new_ids = list(range(start_idx, start_idx + len(objs)))
        else:
            # dict storage path (less efficient but correct)
            new_ids = []
            current_max = max(self._data.keys(), default=-1)
            for i, obj in enumerate(objs):
                idx = current_max + 1 + i
                self._data[idx] = obj
                new_ids.append(idx)
        
        if metadata:
            # metadata is a list of dicts corresponding to objs
            for i, meta in enumerate(metadata):
                if meta:
                    self._meta[new_ids[i]] = meta
                    
        return new_ids

    def set(self, container: StorageType) -> int:
        if not isinstance(container, (list, dict)):
            raise TypeError("container must be a list or a dict[int, Any]")
        self._data = container
        self._meta.clear()
        return len(self._data) - 1 if isinstance(self._data, list) else (max(self._data.keys(), default=-1))

    def remove(self, obj_id: int) -> None:
        try:
            del self._data[obj_id]
        except (IndexError, KeyError):
            raise KeyError(f"No object found with id {obj_id}") from None
        self._meta.pop(obj_id, None)

    def get(self, obj_id: Optional[int] = None):
        if obj_id is None:
            return self._data
        try:
            return self._data[obj_id]
        except (IndexError, KeyError):
            raise KeyError(f"No object found with id {obj_id}") from None

    def list_ids(self) -> List[int]:
        return list(range(len(self._data))) if isinstance(self._data, list) else list(self._data.keys())

    def count(self) -> int:
        return len(self._data)

    def clear(self) -> None:
        if isinstance(self._data, list):
            self._data.clear()
        else:
            self._data = {}
        self._meta.clear()

    # Optional metadata helpers
    def get_meta(self, obj_id: int) -> Dict[str, Any]:
        return dict(self._meta.get(obj_id, {}))

    def set_meta(self, obj_id: int, meta: Dict[str, Any]) -> None:
        self._meta[obj_id] = dict(meta)





import unittest
import pandas as pd
import numpy as np
import sys
import os

# Add parent directory to path
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from feature_analysis import find_discriminative_features

class TestDiscriminativeAnalysis(unittest.TestCase):
    
    def setUp(self):
        np.random.seed(42)
        N = 100
        # Feature A: High for Low Energy (Good), Low for High Energy (Bad)
        # Feature B: Random noise
        
        # Energy: 0 to 1
        y = np.linspace(0, 1, N)
        
        # Feature A correlates negatively with Energy (High A = Low Energy = Good)
        f_a = 1.0 - y + np.random.normal(0, 0.1, N)
        
        # Feature B is random
        f_b = np.random.rand(N)
        
        self.X = pd.DataFrame({'Feat_A': f_a, 'Feat_B': f_b})
        self.y = y

    def test_find_discriminative_features(self):
        results = find_discriminative_features(self.X, self.y, q_lower=0.2, q_upper=0.8)
        
        self.assertFalse(results.empty)
        # Feat_A should have high absolute score
        # Feat_B should be low
        
        score_a = results.loc['Feat_A', 'discriminative_score']
        score_b = results.loc['Feat_B', 'discriminative_score']
        
        # A should be positive (Mean Low (High Val) - Mean High (Low Val)) > 0
        self.assertTrue(score_a > 1.0, f"Score for A should be high, got {score_a}")
        self.assertTrue(abs(score_b) < 1.0, f"Score for B should be low, got {score_b}")
        
    def test_empty_groups_handling(self):
        # Pass quantiles that result in empty groups
        try:
             # q_lower= -0.1 impossible -> empty low group
            res = find_discriminative_features(self.X, self.y, q_lower=-0.1, q_upper=1.1)
            self.assertTrue(res.empty)
        except Exception as e:
            self.fail(f"Function crashed on empty groups: {e}")

if __name__ == '__main__':
    unittest.main()



# =============================
# descriptors_spectral.py
# =============================

import numpy as np
import networkx as nx

class SpectralDescriptors:
    @staticmethod
    def spectral_invariants(G, prefix, k=20):
        n = G.number_of_nodes()
        if n == 0:
            return {f"{prefix}_lambda2": 0, f"{prefix}_spectral_entropy": 0}
        L = nx.laplacian_matrix(G).astype(float).todense()
        evals = np.linalg.eigvalsh(L)
        evals = np.sort(np.maximum(evals, 1e-12))
        if len(evals) > k:
            evals = evals[:k]
        lam2 = evals[1] if len(evals) > 1 else 0
        p = evals / np.sum(evals)
        sent = -np.sum(p * np.log(p))
        return {f"{prefix}_lambda2": float(lam2), f"{prefix}_spectral_entropy": float(sent)}

    @staticmethod
    def local_spectral_entropy(G, prefix, radius=2):
        """
        Calculates stats (mean, std) of spectral entropy for local ego-graphs.
        """
        nodes = list(G.nodes())
        n = len(nodes)
        if n == 0:
            return {}

        entropies = []
        for node in nodes:
            # Extract local ego_graph (radius=radius)
            # Undirected
            sub = nx.ego_graph(G, node, radius=radius)
            if sub.number_of_nodes() < 2:
                entropies.append(0.0)
                continue
            
            # Simple Von Neumann entropy approx on laplacian eigenvalues
            L = nx.laplacian_matrix(sub).astype(float).todense()
            try:
                evals = np.linalg.eigvalsh(L)
                evals = np.sort(np.maximum(evals, 1e-12))
                
                # Normalize density matrix proxy from laplacian?
                # Usually Von Neumann entropy is -Trace(rho log rho). 
                # For graph structural entropy, we use p_i = lambda_i / sum(lambda).
                total_ev = np.sum(evals)
                if total_ev > 1e-9:
                    p = evals / total_ev
                    ent = -np.sum(p * np.log(p))
                    entropies.append(ent)
                else:
                    entropies.append(0.0)
            except:
                entropies.append(0.0)
                
        entropies = np.array(entropies)
        return {
            f"{prefix}_local_spect_entropy_mean": float(np.mean(entropies)),
            f"{prefix}_local_spect_entropy_std": float(np.std(entropies))
        }




from setuptools import setup, find_packages

extras = {
    'analysis': [
        'scipy>=1.7.0',
    ],
    'dev': [
        'pytest>=6.0',  # Specified minimum version for compatibility
        'flake8>=3.8',
        'black',  # Added black for code formatting
        'mypy',  # Added mypy for type checking
        'coverage',  # Added for code coverage analysis
        'tox',  # Added for testing across multiple environments
    ],
    'docs': [
        'sphinx>=4.0',
        'sphinx-rtd-theme>=0.5.0',
        'myst-parser',  # Added MyST parser for enhanced Markdown support in Sphinx
        'sphinx-autodoc-typehints',  # Added for better type hint documentation
    ],
    'ml': [
        'scikit-learn>=0.24.0',
        'umap-learn>=0.5.0',
        'kneed>=0.7.0',
        'hdbscan>=0.8.0',
        'dscribe',
    ],
    'viz': [
        'matplotlib>=3.4.0',
        'seaborn>=0.11.0',
        'imageio>=2.10.0',
        'glfw',
        'PyOpenGL',
        'PyOpenGL_accelerate',
        'imgui[glfw]',
        'Pillow',
    ],
}

# Automatically combine all extras into 'all'
extras['all'] = sorted(list(set(dep for deps in extras.values() for dep in deps)))

setup(
    name='sage_lib',  # El nombre del paquete
    version='0.1.6.65',
    description='A library for advanced scientific calculations and visualization',
    long_description=open('README.md', encoding='utf-8').read(),  # Updated to specify encoding for better compatibility
    long_description_content_type='text/markdown',
    author='Lombardi, Juan Manuel',
    author_email='lombardi@fhi-berlin.mpg.de',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    license='MIT',
    install_requires=[
        'numpy>=1.21.0',  # Updated to a more recent version for better compatibility
        'tqdm>=4.50.0',
        'joblib>=1.0.0',
        'h5py>=3.6.0',
    ],
    extras_require=extras,
    entry_points={
        'console_scripts': [
            'sage = sage_lib.main:main',  # Establece 'sage' como el comando de consola
        ],
    },
    keywords='scientific calculations visualization computational chemistry quantum mechanics machine learning data analysis',  # Expanded keywords for better discoverability
    #project_urls={
    #    'Bug Reports': '',
    #    'Source': '',
    #    'Documentation': '',  # Added a link to documentation
    #},
)



# sage_lib/visualization/browser.py
# ------------------------------------------------------------
# Interactive GUI launcher for structure visualization
# ------------------------------------------------------------
from .PartitionProvider import PartitionProvider
from .visualizer import Viewer


def gui(partition, start_index: int = 0, replicas=(0, 0, 0),
        draw_boxes_for_all: bool = False, verbose: bool = True):
    """
    Launch an interactive GUI visualization for the given Partition object.

    Parameters
    ----------
    partition : Partition
        The Partition object containing atomic structures.
    start_index : int, optional
        Index of the first structure to display. Default is 0.
    replicas : tuple(int, int, int), optional
        Periodic replicas to visualize (nx, ny, nz). Default is (100, 100, 0).
    draw_boxes_for_all : bool, optional
        If True, draw bounding boxes for all structures. Default is False.
    verbose : bool, optional
        If True, print progress and summary messages to the console.

    Behavior
    --------
    Opens an interactive 3D visualization window for exploring atomic structures.
    """
    if verbose:
        print("==============================================")
        print(" Sage-Lib Visualization Interface")
        print("==============================================")
        print("Initializing PartitionProvider...")

    # Initialize provider
    provider = PartitionProvider(partition)
    n_structs = len(provider)
    print( partition.get_all_energies().shape )
    if verbose:
        print(f" → Loaded Partition with {n_structs} structures.")
        print(f" → Starting visualization from index {start_index}.")
        print(f" → Periodic replicas: {replicas}")
        print(f" → Bounding boxes for all: {draw_boxes_for_all}")
        print("----------------------------------------------")
        print("Launching visualization window...")
        print("Use arrow keys or navigation controls to browse.")
        print("Close the window to return to terminal.")
        print("==============================================")

    # Launch the visualization
    '''
    browse_structures(
        provider,
        start_index=start_index,
        replicas=replicas,
        draw_boxes_for_all=draw_boxes_for_all
    )
    '''
    viewer = Viewer(provider)
    viewer.loop()

    if verbose:
        print("\nVisualization session ended.")
        print(f"Displayed {n_structs} structures in total.")
        print("Returning to CLI context.\n")



# =============================
# descriptors_topology.py
# =============================

import numpy as np
import networkx as nx
from scipy.stats import skew

class TopologyDescriptors:
    @staticmethod
    def degree_stats(G, prefix):
        deg = np.array([d for _, d in G.degree()], float)
        if len(deg) == 0:
            return {f"{prefix}_deg_mean": 0, f"{prefix}_deg_std": 0, f"{prefix}_deg_skew": 0}

        return {
            f"{prefix}_deg_mean": float(np.mean(deg)),
            f"{prefix}_deg_std": float(np.std(deg)),
            f"{prefix}_deg_skew": float(skew(deg))
        }

    @staticmethod
    def degree_stats_by_species(G, prefix):
        """
        Mean degree separated by element species (e.g. deg_mean_Ni).
        """
        if G.number_of_nodes() == 0:
            return {}

        stats = {}
        species_degrees = {}

        for n, d in G.degree():
            el = G.nodes[n]["element"]
            species_degrees.setdefault(el, []).append(d)

        for el, degs in species_degrees.items():
            stats[f"{prefix}_deg_mean_{el}"] = float(np.mean(degs))
            stats[f"{prefix}_deg_std_{el}"] = float(np.std(degs))
        
        return stats

    @staticmethod
    def cycle_stats(G, prefix):
        try:
            cycles = nx.cycle_basis(G)
        except Exception:
            cycles = []
        if not cycles:
            return {f"{prefix}_n_cycles": 0, f"{prefix}_cycle_mean": 0}
        lengths = [len(c) for c in cycles]
        return {
            f"{prefix}_n_cycles": len(cycles),
            f"{prefix}_cycle_mean": float(np.mean(lengths))
        }

    @staticmethod
    def clustering(G, prefix):
        if G.number_of_nodes() == 0:
            return {f"{prefix}_clust": 0}
        return {f"{prefix}_clust": nx.average_clustering(G, weight="weight")}

    @staticmethod
    def path_stats(G, prefix):
        if G.number_of_nodes() == 0:
            return {f"{prefix}_path_mean": 0, f"{prefix}_path_std": 0}
        comp = max(nx.connected_components(G), key=len)
        H = G.subgraph(comp)
        if H.number_of_nodes() <= 1:
            return {f"{prefix}_path_mean": 0, f"{prefix}_path_std": 0}
        sp = dict(nx.all_pairs_shortest_path_length(H))
        dist = []
        for i in sp:
            dist += list(sp[i].values())
        dist = np.array(dist, float)
        return {f"{prefix}_path_mean": float(np.mean(dist)), f"{prefix}_path_std": float(np.std(dist))}




import math
import numpy as np
from typing import Dict, List, Sequence, Tuple
from scipy.spatial import Voronoi
from collections.abc import Iterable
from collections import Counter
try:
    from .BasePartition import BasePartition
except ImportError as e:
    import sys
    sys.stderr.write(f"An error occurred while importing numpy: {str(e)}\n")
    del sys

class Crystal_builder(BasePartition):
    """
    A class for generating crystalline structures.

    This class provides methods to generate various types of crystal cells,
    including but not limited to cubic, face-centered cubic (FCC), body-centered
    cubic (BCC), tetragonal, body-centered tetragonal (BCT), orthorhombic,
    hexagonal, rhombohedral, monoclinic, triclinic as well as 2D lattices such as
    square, rectangular, oblique, and hexagonal (2D). The lattice vectors and atomic
    positions are computed based on the selected lattice type and provided parameters.
    The resulting structure is used to populate the AtomPositionManager attributes
    within a container.

    Attributes
    ----------
    lattice_type : str
        Identifier for the crystal lattice type (e.g. 'cubic', 'fcc', 'bcc',
        'tetragonal', 'bct', 'orthorhombic', 'hexagonal', 'rhombohedral',
        'monoclinic', 'triclinic', 'square2d', 'rectangular2d', 'oblique2d', 'hexagonal2d').
    lattice_parameters : dict
        Dictionary holding the parameters specific to the chosen lattice type.
    atom_positions : numpy.ndarray
        Array containing the atomic positions within the unit cell.
    lattice_vectors : numpy.ndarray
        3x3 array representing the lattice vectors.
    atom_labels : list
        List of atom labels for each atom in the structure.
    atom_count : int
        Total number of atoms in the unit cell.
    pbc : list
        Periodic boundary conditions in each direction (typically [True, True, True]).

    Methods
    -------
    set_lattice_type(lattice_type: str, **params)
        Set the lattice type and its required parameters.
    generate_lattice_vectors()
        Compute the lattice vectors based on the chosen lattice type.
    generate_atom_positions(atom_label: str = 'H')
        Generate the atomic positions for the cell.
    build_crystal(lattice_type: str, lattice_parameters: dict)
        Build the complete crystal structure and populate the AtomPositionManager.
    """
    SUPPORTED_LATTICES: Tuple[str, ...] = (
        "cubic", "fcc", "bcc",
        "tetragonal", "bct",
        "orthorhombic", "hexagonal",
        "rhombohedral", "monoclinic", "triclinic",
        #"square2d", "rectangular2d", "oblique2d", "hexagonal2d",
    )

    def __init__(self, *args, **kwargs):
        """
        """

        # Initialize lattice-related attributes with default values
        self.lattice_type = None
        self.lattice_parameters = {}
        self.atom_positions = None
        self.lattice_vectors = None
        self.atom_labels = []
        self.atom_count = 0
        self.pbc = [True, True, True]
        self._primitive_basis_cart: np.ndarray | None = None  # new
        
        self.hex_c_over_a = 1.633  # ideal HCP ratio; used by hexagonal branch

        super().__init__(*args, **kwargs)
    
    # ==================================================================
    #  Helpers – geometry & hashing
    # ==================================================================
    @staticmethod
    def _structure_signature(cart: np.ndarray, *, decimals: int = 3) -> Tuple[float, ...]:
        """Order‑independent tuple of rounded pairwise distances."""
        n = len(cart)
        dists: List[float] = []
        for i in range(n):
            for j in range(i + 1, n):
                d = np.linalg.norm(cart[i] - cart[j])
                dists.append(round(d, decimals))
        return tuple(sorted(dists))

    def _dist_pbc(self, p: np.ndarray, q: np.ndarray, min_distance: float) -> bool:
        """Return *True* if ‖p−q‖ (Cartesian, PBC‑aware) > *min_distance*."""
        d = p - q
        d -= np.rint(d)  # wrap into [-0.5, 0.5]
        cart = d @ self.lattice_vectors
        return np.linalg.norm(cart) > min_distance

    def _unique_rows(self, arr: np.ndarray, tol: float = 1e-6) -> np.ndarray:
        out: List[np.ndarray] = []
        for r in arr:
            if not any(np.allclose(r, x, atol=tol) for x in out):
                out.append(r)
        return np.asarray(out)

    # ==================================================================
    #  Volume-per-atom helpers (new)
    # ==================================================================
    def _packing_fraction_default(self, lt: str) -> float:
        """Close-packing fraction φ* used to convert sphere volumes → per-atom cell volume."""
        lt = lt.lower()
        if lt in {"fcc"}:
            return 0.74048        # FCC
        if lt in {"hexagonal", "hcp"}:
            return 0.74048        # HCP (ideal c/a)
        if lt in {"bcc", "bct"}:
            return 0.68017        # BCC/BCT
        if lt in {"cubic", "sc", "tetragonal", "orthorhombic"}:
            return math.pi / 6.0  # ≈ 0.5236 (simple cubic baseline)
        # Rhombohedral/Monoclinic/Triclinic vary with angles; keep a conservative default
        return 0.60

    def _mean_atomic_sphere_volume(self, species: Sequence[str]) -> float:
        """Average per-atom sphere volume from covalent radii."""
        vols = [(4.0/3.0) * math.pi * (self.covalent_radii[s] ** 3) for s in species]
        return float(np.mean(vols))

    def _cell_volume_per_atom(self, species: Sequence[str], lt: str, packing: float) -> float:
        """Target per-atom cell volume: V* = (⟨V_sphere⟩ / φ*) × packing."""
        phi = self._packing_fraction_default(lt)
        return (self._mean_atomic_sphere_volume(species) / phi) * float(packing)

    def _a_from_target_volume(self, lt: str, V_atom: float, angles: Tuple[float, float, float]) -> float:
        """Solve for length 'a' such that det(L(a,angles)) == V_atom (one atom / primitive cell)."""
        alpha, beta, gamma = angles

        # Build the lattice with a=1 to measure the geometric factor F = det(L(a=1))
        backup_lv = None if self.lattice_vectors is None else self.lattice_vectors.copy()
        self._prepare_lattice(lt, 1.0, alpha, beta, gamma)
        F = abs(np.linalg.det(self.lattice_vectors))
        if backup_lv is not None:
            self.lattice_vectors = backup_lv
        if F <= 0:
            raise RuntimeError(f"Degenerate lattice for {lt} with angles={angles}: det=0")

        # Volume scales as a^3 for all branches defined in _prepare_lattice
        return (V_atom / F) ** (1.0 / 3.0)

    def set_lattice_type(self, lattice_type: str, **params):
        """
        Set the lattice type and its associated parameters.

        Parameters
        ----------
        lattice_type : str
            Identifier for the crystal lattice type. Supported types include:
            'cubic', 'fcc', 'bcc', 'tetragonal'.
        **params
            Lattice-specific parameters. For example, for a cubic lattice provide 'a',
            for a tetragonal cell provide 'a' and 'c', etc.
        """
        self.lattice_type = lattice_type.lower()
        self.lattice_parameters = params

    def generate_lattice_vectors(self):
        """
        Compute lattice vectors based on the selected lattice type and parameters.

        Returns
        -------
        numpy.ndarray
            A 3x3 array containing the lattice vectors.

        Raises
        ------
        ValueError
            If the lattice type is not set or required parameters are missing.
        """
        if self.lattice_type is None:
            raise ValueError("Lattice type not set. Use set_lattice_type() first.")

        lp = self.lattice_parameters  # shorthand
        # 3D lattices
        if self.lattice_type == 'cubic':
            try:
                a = float(lp['a'])
            except KeyError:
                raise ValueError("Parameter 'a' must be provided for cubic lattice.")
            self.lattice_vectors = a * np.eye(3)

        elif self.lattice_type == 'fcc':
            try:
                a = float(lp['a'])
            except KeyError:
                raise ValueError("Parameter 'a' must be provided for FCC lattice.")
            self.lattice_vectors = 0.5 * np.array([[0, a, a],
                                                    [a, 0, a],
                                                    [a, a, 0]])

        elif self.lattice_type == 'bcc':
            try:
                a = float(lp['a'])
            except KeyError:
                raise ValueError("Parameter 'a' must be provided for BCC lattice.")
            self.lattice_vectors = 0.5 * np.array([[-a, a, a],
                                                    [a, -a, a],
                                                    [a, a, -a]])

        elif self.lattice_type == 'tetragonal':
            try:
                a = float(lp['a'])
                c = float(lp['c'])
            except KeyError:
                raise ValueError("Parameters 'a' and 'c' must be provided for tetragonal lattice.")
            self.lattice_vectors = np.diag([a, a, c])

        elif self.lattice_type == 'bct':
            try:
                a = float(lp['a'])
                c = float(lp['c'])
            except KeyError:
                raise ValueError("Parameters 'a' and 'c' must be provided for body-centered tetragonal lattice.")
            # Primitive cell for body-centered tetragonal (analogous to BCC with tetragonal distortion)
            self.lattice_vectors = 0.5 * np.array([[ a,  a,  c],
                                                    [-a,  a,  c],
                                                    [ a, -a,  c]])

        elif self.lattice_type == 'orthorhombic':
            try:
                a = float(lp['a'])
                b = float(lp['b'])
                c = float(lp['c'])
            except KeyError:
                raise ValueError("Parameters 'a', 'b', and 'c' must be provided for orthorhombic lattice.")
            self.lattice_vectors = np.diag([a, b, c])

        elif self.lattice_type == 'hexagonal':
            try:
                a = float(lp['a'])
                c = float(lp['c'])
            except KeyError:
                raise ValueError("Parameters 'a' and 'c' must be provided for hexagonal lattice.")
            self.lattice_vectors = np.array([[ a,        0, 0],
                                             [-a/2, a*np.sqrt(3)/2, 0],
                                             [ 0,        0, c]])

        elif self.lattice_type == 'rhombohedral':
            try:
                a = float(lp['a'])
                alpha = float(lp['alpha'])
            except KeyError:
                raise ValueError("Parameters 'a' and 'alpha' must be provided for rhombohedral lattice.")
            alpha_rad = np.deg2rad(alpha)
            v1 = np.array([a, 0, 0])
            v2 = np.array([a * np.cos(alpha_rad), a * np.sin(alpha_rad), 0])
            # A common representation for a rhombohedral cell:
            # v3 is defined so that the angle between any two vectors equals alpha.
            v3_x = a * np.cos(alpha_rad)
            # Using an equivalent formulation: v3_y = a*(1 - cos(alpha))/tan(alpha)
            v3_y = a * (1 - np.cos(alpha_rad)) / np.tan(alpha_rad) if np.tan(alpha_rad) != 0 else 0
            v3_z = a * np.sqrt(1 - 3*np.cos(alpha_rad)**2 + 2*np.cos(alpha_rad)**3)
            v3 = np.array([v3_x, v3_y, v3_z])
            self.lattice_vectors = np.array([v1, v2, v3])

        elif self.lattice_type == 'monoclinic':
            try:
                a = float(lp['a'])
                b = float(lp['b'])
                c = float(lp['c'])
                beta = float(lp['beta'])
            except KeyError:
                raise ValueError("Parameters 'a', 'b', 'c', and 'beta' must be provided for monoclinic lattice.")
            beta_rad = np.deg2rad(beta)
            v1 = np.array([a, 0, 0])
            v2 = np.array([0, b, 0])
            v3 = np.array([c * np.cos(beta_rad), 0, c * np.sin(beta_rad)])
            self.lattice_vectors = np.array([v1, v2, v3])

        elif self.lattice_type == 'triclinic':
            try:
                a = float(lp['a'])
                b = float(lp['b'])
                c = float(lp['c'])
                alpha = float(lp['alpha'])
                beta  = float(lp['beta'])
                gamma = float(lp['gamma'])
            except KeyError:
                raise ValueError("Parameters 'a', 'b', 'c', 'alpha', 'beta', and 'gamma' must be provided for triclinic lattice.")
            alpha_rad = np.deg2rad(alpha)
            beta_rad  = np.deg2rad(beta)
            gamma_rad = np.deg2rad(gamma)
            v1 = np.array([a, 0, 0])
            v2 = np.array([b * np.cos(gamma_rad), b * np.sin(gamma_rad), 0])
            v3_x = c * np.cos(beta_rad)
            v3_y = c * (np.cos(alpha_rad) - np.cos(beta_rad) * np.cos(gamma_rad)) / np.sin(gamma_rad)
            term = 1 - np.cos(beta_rad)**2 - ((np.cos(alpha_rad) - np.cos(beta_rad)*np.cos(gamma_rad)) / np.sin(gamma_rad))**2
            if term < 0:
                term = 0
            v3_z = c * np.sqrt(term)
            v3 = np.array([v3_x, v3_y, v3_z])
            self.lattice_vectors = np.array([v1, v2, v3])

        # 2D lattices (assume the third row is zero)
        elif self.lattice_type == 'square2d':
            try:
                a = float(lp['a'])
            except KeyError:
                raise ValueError("Parameter 'a' must be provided for square2d lattice.")
            self.lattice_vectors = np.array([[a, 0, 0],
                                             [0, a, 0],
                                             [0, 0, 0]])
        elif self.lattice_type == 'rectangular2d':
            try:
                a = float(lp['a'])
                b = float(lp['b'])
            except KeyError:
                raise ValueError("Parameters 'a' and 'b' must be provided for rectangular2d lattice.")
            self.lattice_vectors = np.array([[a, 0, 0],
                                             [0, b, 0],
                                             [0, 0, 0]])
        elif self.lattice_type == 'oblique2d':
            try:
                a = float(lp['a'])
                b = float(lp['b'])
                gamma = float(lp['gamma'])
            except KeyError:
                raise ValueError("Parameters 'a', 'b', and 'gamma' must be provided for oblique2d lattice.")
            gamma_rad = np.deg2rad(gamma)
            self.lattice_vectors = np.array([[a, 0, 0],
                                             [b * np.cos(gamma_rad), b * np.sin(gamma_rad), 0],
                                             [0, 0, 0]])
        elif self.lattice_type == 'hexagonal2d':
            try:
                a = float(lp['a'])
            except KeyError:
                raise ValueError("Parameter 'a' must be provided for hexagonal2d lattice.")
            self.lattice_vectors = np.array([[a, 0, 0],
                                             [a/2, a*np.sqrt(3)/2, 0],
                                             [0, 0, 0]])
        else:
            raise ValueError(f"Lattice type '{self.lattice_type}' not supported.")

        return self.lattice_vectors

    def generate_atom_positions(self, atom_label: str = 'X'):
        """
        Generate atomic positions for the unit cell based on the lattice type.

        Returns
        -------
        numpy.ndarray
            An array of atomic positions.

        Notes
        -----
        For primitive cells a single atom is placed at the origin. For lattices with a basis,
        such as FCC, BCC, and BCT, multiple atomic positions are provided. In 2D systems the z-coordinate
        is set to zero.
        """
        if self.lattice_type is None:
            raise ValueError("Lattice type not set. Use set_lattice_type() first.")

        # Ensure lattice vectors are generated
        if self.lattice_vectors is None:
            self.generate_lattice_vectors()

        lt = self.lattice_type

        if lt == "cubic":
            basis_frac = np.array([[0.0, 0.0, 0.0]])
        elif lt == "fcc":
            basis_frac = np.array([
                [0.0, 0.0, 0.0],
                [0.0, 0.5, 0.5],
                [0.5, 0.0, 0.5],
                [0.5, 0.5, 0.0],
            ])
        elif lt == "bcc":
            basis_frac = np.array([
                [0.0, 0.0, 0.0],
                [0.5, 0.5, 0.5],
            ])
        elif lt == "bct":
            basis_frac = np.array([
                [0.0, 0.0, 0.0],
                [0.5, 0.5, 0.5],
            ])
        elif lt in (
            "tetragonal",
            "orthorhombic",
            "hexagonal",
            "rhombohedral",
            "monoclinic",
            "triclinic",
        ):
            basis_frac = np.array([[0.0, 0.0, 0.0]])
        # -------- 2D lattices (z = 0) ----------------------------------
        elif lt == "square2d":
            basis_frac = np.array([[0.0, 0.0, 0.0]])
        elif lt == "rectangular2d":
            basis_frac = np.array([[0.0, 0.0, 0.0]])
        elif lt == "oblique2d":
            basis_frac = np.array([[0.0, 0.0, 0.0]])
        elif lt == "hexagonal2d":
            basis_frac = np.array([
                [0.0, 0.0, 0.0],
                [2.0 / 3.0, 1.0 / 3.0, 0.0],
            ])
        else:
            raise ValueError(f"Atom positions for lattice '{lt}' not implemented.")

        # Cartesian positions
        self._primitive_basis_cart = basis_frac @ self.lattice_vectors
        self.atom_positions = self._primitive_basis_cart.copy()
        self.atom_labels = [atom_label] * len(self.atom_positions)
        self.atom_count = len(self.atom_positions)
        return self.atom_positions

    def build_crystal(self, lattice_type:str='fcc', lattice_parameters:dict={'a':1}, atom_label:str='H'):
        """
        Build the crystal structure and populate the AtomPositionManager attributes.

        This method calculates the lattice vectors and atomic positions using the
        specified lattice type and parameters. The generated structure data can then
        be assigned to the container's AtomPositionManager.

        Returns
        -------
        dict
            A dictionary with keys:
            - 'atomCount': number of atoms in the cell.
            - 'atomPositions': numpy array of atomic positions.
            - 'atomLabelsList': list of atomic labels.
            - 'latticeVectors': 3x3 array of lattice vectors.
            - 'pbc': periodic boundary conditions list.
        
        Example
        -------
        >>> container = SingleRun('')
        >>> cpb = Crystal_builder()
        >>> cpb.set_lattice_type('fcc', a=4.0)
        >>> structure = cpb.build_crystal()
        >>> container.AtomPositionManager.atomCount = structure['atomCount']
        >>> container.AtomPositionManager.atomPositions = structure['atomPositions']
        >>> container.AtomPositionManager.atomLabelsList = structure['atomLabelsList']
        >>> container.AtomPositionManager.latticeVectors = structure['latticeVectors']
        >>> container.AtomPositionManager.pbc = structure['pbc']
        >>> self.add_container(container)
        """

        container = self.add_empty_container()
        lattice_parameters = lattice_parameters or {"a": 1.0}

        self.set_lattice_type(lattice_type, 
            a=lattice_parameters.get('a', 0), 
            b=lattice_parameters.get('b', 0), 
            c=lattice_parameters.get('c', 0),
            alpha=lattice_parameters.get('alpha', 0),
            beta=lattice_parameters.get('beta', 0),
            gamma=lattice_parameters.get('gamma', 0) )

        self.generate_lattice_vectors()
        self.generate_atom_positions(atom_label=atom_label)

        structure_data = {
            'atomCount': self.atom_count,
            'atomPositions': self.atom_positions,
            'atomLabelsList': self.atom_labels,
            'latticeVectors': self.lattice_vectors,
            'pbc': self.pbc
        }   

        container.AtomPositionManager.set_latticeVectors( structure_data['latticeVectors'] )
        container.AtomPositionManager.add_atom(structure_data['atomLabelsList'], structure_data['atomPositions'])
        self.add_container(container)

        return container

    # ------------------------------------------------------------------
    #  Public API – single cell via Voronoi
    # ------------------------------------------------------------------
    def build_crystal_voronoi(
        self,
        lattice_type: str,
        species: List[str],
        packing: float = 1.15,   # tighter default recommended
        pad: int = 1,
    ):
        angles = (90.0, 90.0, 90.0)
        V_atom = self._cell_volume_per_atom(species, lattice_type, packing)
        a_len  = self._a_from_target_volume(lattice_type, V_atom, angles)
        self._prepare_lattice(lattice_type, a_len, *angles)

        sites_frac = self._voronoi_vertices(pad)
        cart = self._select_sites_cartesian(sites_frac, len(species))
        return self._populate_container(species, cart)

    # ------------------------------------------------------------------
    #  Public API – recursive list (k = 1 … N)
    # ------------------------------------------------------------------
    def build_all_voronoi_cells(
        self,
        lattice_type: str,
        species_list: List[str],
        packing: float = 1.15,   # tighter default recommended
        pad: int = 1,
    ) -> Dict[int, "BasePartition"]:
        N = len(species_list)
        angles = (90.0, 90.0, 90.0)
        V_atom = self._cell_volume_per_atom(species_list, lattice_type, packing)
        a_len  = self._a_from_target_volume(lattice_type, V_atom, angles)
        self._prepare_lattice(lattice_type, a_len, *angles)

        sites_frac = self._voronoi_vertices(pad)
        result: Dict[int, BasePartition] = {}

        def _recurse(k: int):
            if k == 0:
                return
            cart = self._select_sites_cartesian(sites_frac, k)
            cont = self._populate_container(species_list[:k], cart)
            result[k] = cont
            _recurse(k - 1)

        _recurse(N)
        return result

    # ==================================================================
    #  Public API – exhaustive enumeration with duplicate culling
    # ==================================================================
    def build_all_crystals(
        self,
        species_list: List[str],
        packing: float = 2.25,
        pad: int = 1,
        angle_delta: float = 0.0,
        *,
        min_distance: float = 1e-3,
        dedup_tol: float = 1e-1,
        max_vertices: int | None = 2,
        max_structures: int | None = None,
    ) -> Dict[Tuple[str, int, Tuple[float, float, float], int], "BasePartition"]:
        """See original doc‑string; new keyword‑only knobs described below.

        | Parameter            | Type / Unit | Default | Meaning                                                                                                                                                                                                               |                                                                                                                                                          |
        | -------------------- | ----------- | ------- | --------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- | -------------------------------------------------------------------------------------------------------------------------------------------------------- |
        | **`species_list`**   | `list[str]` | –       | Multiset of chemical symbols to distribute (e.g. `['Fe','O','O']`). Every permutation is explored.                                                                                                                    |                                                                                                                                                          |
        | **`packing`**        | `float`     | `1.25`  | Safety factor applied to the *isotropic* cell length `a = packing × Σ(2 rᵢ)` where `rᵢ` are the covalent radii of the atoms already placed. Larger values ⇒ more spacious cells.                                      |                                                                                                                                                          |
        | **`pad`**            | `int`       | `1`     | How far the periodic image grid extends (± *pad* replicas in each lattice direction) when computing Voronoi vertices.                                                                                                 |                                                                                                                                                          |
        | **`angle_delta`**    | `float` (°) | `0.0`   | Generates a small grid of cell angles around 90°. If 0 ⇒ only (90, 90, 90). If *d* > 0 ⇒ Cartesian product (90 ± *d*)³ (max 27 angle triplets).                                                                       |                                                                                                                                                          |
        | **`min_distance`**   | `float` (Å) | `1e-3`  | Hard lower bound for every *Cartesian* pairwise separation. A default of 0.001 Å is practically “disabled”, preserving legacy behaviour unless you specify a larger value.                                            |                                                                                                                                                          |
        | **`dedup_tol`**      | `float` (Å) | `1e-3`  | Resolution used when rounding the inter-atomic distances that form the **structure signature**. Two signatures identical to this tolerance are deemed geometrically equivalent and the second structure is discarded. |                                                                                                                                                          |
        | **`max_vertices`**   | `int        |  None`  | `2`                                                                                                                                                                                                                   | Optional cap on the number of Voronoi vertices tested **per insertion step**. `None` keeps all vertices; a small integer throttles the branching factor. |
        | **`max_structures`** | `int        |  None`  | `None`                                                                                                                                                                                                                | Global ceiling on the number of *accepted, unique* structures. When the limit is reached the search terminates immediately.                              |

        Parameters (additions)
        ---------------------
        min_distance : float, default 1e‑3 Å
            Hard lower bound for every pairwise separation.
        dedup_tol : float, default 1e‑3 Å
            Tolerance used when rounding distances for duplicate removal.
        max_structures : int | None
            If given, stop once this many unique geometries have been accepted.
        """
        class _MaxReached(Exception):
            """Private sentinel raised when the hard ceiling is hit."""
            pass

        if not species_list:
            raise ValueError("species_list must contain at least one element.")

        # --------------- angle grid -----------------------------------
        if angle_delta <= 0.0:
            angle_grid = [(90.0, 90.0, 90.0)]
        else:
            offs = (-angle_delta, 0.0, angle_delta)
            angle_grid = [(90.0 + da, 90.0 + db, 90.0 + dc)
                          for da in offs for db in offs for dc in offs]

        # --------------- bookkeeping ----------------------------------
        result: Dict[Tuple[str, int, Tuple[float, float, float], int], BasePartition] = {}
        signatures: set[Tuple[float, ...]] = set()
        branch_id = 0
        decimals = max(0, int(round(-math.log10(dedup_tol)))) if dedup_tol > 0 else 3

        def register(cont: "BasePartition", cart: np.ndarray,
                      key: Tuple[str, int, Tuple[float, float, float], int]):
            sig = self._structure_signature(cart, decimals=decimals)
            if sig in signatures:
                return  # duplicate – ignore
            signatures.add(sig)
            result[key] = cont
            if max_structures is not None and len(result) >= max_structures:
                raise _MaxReached

        # --------------- DFS ------------------------------------------
        def dfs(atoms_frac: List[np.ndarray], remaining: Counter, labels_order: List[str],
                lt: str, angles: Tuple[float, float, float]):
            nonlocal branch_id
            depth = len(labels_order)

            if depth:  # store prefix
                cart = np.vstack(atoms_frac) @ self.lattice_vectors
                cont = self._populate_container(labels_order, cart)
                key = (lt, depth, angles, branch_id)
                register(cont, cart, key)

            if not remaining:
                branch_id += 1
                return

            for sp in list(remaining):
                remaining[sp] -= 1
                if remaining[sp] == 0:
                    del remaining[sp]

                new_labels = labels_order + [sp]
                alpha, beta, gamma = angles
                V_atom = self._cell_volume_per_atom(new_labels, lt, packing)
                a_len  = self._a_from_target_volume(lt, V_atom, angles)
                self._prepare_lattice(lt, a_len, alpha, beta, gamma)

                verts = (self._voronoi_vertices_atoms([], pad) if not atoms_frac
                         else self._voronoi_vertices_atoms(atoms_frac, pad))
                good = [v for v in verts if all(
                    self._dist_pbc(v, a, min_distance) for a in atoms_frac)]
                if max_vertices is not None:
                    good = good[:max_vertices]

                if not good:
                    branch_id += 1
                else:
                    for v in good:
                        dfs(atoms_frac + [v], remaining, new_labels, lt, angles)

                remaining[sp] += 1

        # --------------- outer loops -----------------------------------
        try:
            for angles in angle_grid:
                alpha, beta, gamma = angles  # not used directly but kept for clarity
                for lt in self.SUPPORTED_LATTICES:
                    dfs([], Counter(species_list), [], lt, angles)
        except _MaxReached:
            pass  # hard ceiling hit – silent exit

        return result

    # =====================================================================
    #  Internal helpers – geometry, lattice, etc.
    # =====================================================================
    def _largest_diameter(self, species: Sequence[str]) -> float:
        try:
            return 2.0 * max(self.covalent_radii[s] for s in species)
        except KeyError as exc:
            raise ValueError(
                f"Atomic radius for element '{exc.args[0]}' is missing in self._covalent_radii."
            )

    def _total_diameter(self, species: Sequence[str]) -> float:
        """Sum of *individual* diameters (2·r) – generous upper bound."""
        try:
            return sum(2.0 * self.covalent_radii[s] for s in species)
        except KeyError as exc:
            raise ValueError(f"Missing covalent radius for element {exc.args[0]}")

    def _compute_isotropic_size(self, species: Sequence[str], packing: float) -> float:
        return packing * self._total_diameter(species)

    # ------------------------------------------------------------------
    #  Geometry helpers – Voronoi
    # ------------------------------------------------------------------
    def _voronoi_vertices(self, pad: int = 1, *, joggle: bool = True) -> np.ndarray:
        """Return Voronoi‑vertex fractional coordinates of the Wigner–Seitz cell.

        Robust to numerical degeneracies:
        * Uses Qhull *QJ* joggle if needed.
        * Falls back to Moore–Penrose pseudo‑inverse if the lattice matrix is
          (nearly) singular (e.g. 2D lattices embedded in 3D).
        """
        if Voronoi is None:
            raise ModuleNotFoundError(
                "scipy is required for Voronoi‑based site generation – please install scipy"
            )

        rng = range(-pad, pad + 1)
        grid = np.array([[i, j, k] for i in rng for j in rng for k in rng], dtype=float)
        pts_cart = grid @ self.lattice_vectors
        try:
            v = Voronoi(pts_cart)
        except Exception:
            if not joggle:
                raise
            from scipy.spatial import Voronoi as _Voronoi
            v = _Voronoi(pts_cart, qhull_options="Qbb Qc Qx QJ")

        idx0 = len(pts_cart) // 2
        region_idx = v.point_region[idx0]
        vert_idx = v.regions[region_idx]
        if -1 in vert_idx:
            raise RuntimeError("Increase pad; Voronoi region unbounded.")
        verts_cart = v.vertices[vert_idx]
        inv_lv = np.linalg.pinv(self.lattice_vectors)  # robust
        verts_frac = (verts_cart @ inv_lv) % 1.0
        verts_frac = self._unique_rows(verts_frac)
        order = np.argsort(np.sum(verts_frac ** 2, axis=1))
        return verts_frac[order]

    # ---------------- Voronoi of current atoms ------------------------
    def _voronoi_vertices_atoms(self, atoms_frac: Sequence[np.ndarray], pad: int) -> List[np.ndarray]:
        # --- early exit: no atoms yet -> use bare lattice vertices
        if len(atoms_frac) == 0:
            return list(self._voronoi_vertices(pad))

        if Voronoi is None:
            raise ModuleNotFoundError("SciPy required for Voronoi tessellation")
        
        atoms_frac = np.asarray(atoms_frac, dtype=float).reshape(-1, 3)
        #atoms_frac = np.asarray(atoms_frac)
        translations = np.array([[i, j, k] for i in range(-pad, pad + 1)
                                           for j in range(-pad, pad + 1)
                                           for k in range(-pad, pad + 1)], float)
        tiled = atoms_frac[:, None, :] + translations[None, :, :]
        pts_cart = tiled.reshape(-1, 3) @ self.lattice_vectors

        # Voronoi computation with joggle fallback
        try:
            v = Voronoi(pts_cart)
        except Exception:
            from scipy.spatial import Voronoi as _Voronoi
            v = _Voronoi(pts_cart, qhull_options="Qbb Qc Qx QJ")

        verts_cart = v.vertices
        inv_lv = np.linalg.pinv(self.lattice_vectors)
        verts_frac = (verts_cart @ inv_lv) % 1.0
        # deduplicate rows
        unique: List[np.ndarray] = []
        for r in verts_frac:
            if not any(np.allclose(r, u, atol=1e-6) for u in unique):
                unique.append(r)
        return unique

    # ------------------------------------------------------------------
    #  Lattice preparation – currently only primitive cubic (sc) handled
    # ------------------------------------------------------------------
    def _prepare_lattice(self, lt: str, a: float, alpha: float, beta: float, gamma: float):
        lt = lt.lower()
        self.lattice_type = lt

        if lt in {"cubic", "sc"}:  # simple cubic
            self.lattice_vectors = a * np.eye(3)

        elif lt == "fcc":  # face‑centred cubic (primitive)
            self.lattice_vectors = 0.5 * a * np.array([
                [0, 1, 1],
                [1, 0, 1],
                [1, 1, 0],
            ])

        elif lt == "bcc":  # body‑centred cubic (primitive)
            self.lattice_vectors = 0.5 * a * np.array([
                [-1, 1, 1],
                [1, -1, 1],
                [1, 1, -1],
            ])

        elif lt == "tetragonal":  # a = b ≠ c
            c = a  # placeholder – could scale differently
            self.lattice_vectors = np.diag([a, a, c])

        elif lt == "bct":  # body‑centred tetragonal (primitive)
            c = a  # placeholder
            self.lattice_vectors = 0.5 * np.array([
                [ a,  a,  c],
                [-a,  a,  c],
                [ a, -a,  c],
            ])

        elif lt == "orthorhombic":
            b = c = a  # placeholders
            self.lattice_vectors = np.diag([a, b, c])

        elif lt == "hexagonal":
            # Allow ideal c/a unless overridden elsewhere
            c_over_a = getattr(self, "hex_c_over_a", 1.633)  # ideal ≈ sqrt(8/3)
            c = c_over_a * a
            self.lattice_vectors = np.array([
                [ a,        0, 0],
                [-a/2, a*np.sqrt(3)/2, 0],
                [ 0,        0, c],
            ])

        elif lt == "rhombohedral":
            alpha_rad = math.radians(alpha)
            self.lattice_vectors = a * np.array([
                [ 1,                                0,                              0],
                [ math.cos(alpha_rad), math.sin(alpha_rad),                          0],
                [ math.cos(alpha_rad), (math.cos(alpha_rad) - math.cos(alpha_rad)**2)/math.sin(alpha_rad) if abs(math.sin(alpha_rad))>1e-8 else 0,
                  math.sqrt(1 - 3*math.cos(alpha_rad)**2 + 2*math.cos(alpha_rad)**3) ],
            ])

        elif lt == "monoclinic":
            beta_rad = math.radians(beta)
            b = c = a
            self.lattice_vectors = np.array([
                [a, 0, 0],
                [0, b, 0],
                [c*math.cos(beta_rad), 0, c*math.sin(beta_rad)],
            ])

        elif lt == "triclinic":
            alpha_rad = math.radians(alpha)
            beta_rad  = math.radians(beta)
            gamma_rad = math.radians(gamma)
            b = c = a
            v1 = np.array([a, 0, 0])
            v2 = np.array([b*math.cos(gamma_rad), b*math.sin(gamma_rad), 0])
            v3_x = c*math.cos(beta_rad)
            v3_y = c*(math.cos(alpha_rad) - math.cos(beta_rad)*math.cos(gamma_rad)) / max(math.sin(gamma_rad), 1e-8)
            term = 1 - math.cos(beta_rad)**2 - ((math.cos(alpha_rad) - math.cos(beta_rad)*math.cos(gamma_rad)) / max(math.sin(gamma_rad),1e-8))**2
            v3_z = c*math.sqrt(max(term, 0))
            self.lattice_vectors = np.array([v1, v2, [v3_x, v3_y, v3_z]])

        else:
            raise NotImplementedError(f"Lattice '{lt}' not supported.")

        # Warn if user gave non‑90° angles for lattices that ignore them
        if lt in {"cubic", "sc", "fcc", "bcc", "tetragonal", "bct", "orthorhombic", "hexagonal"}:
            if any(abs(x - 90.0) > 1e-2 for x in (alpha, beta, gamma)):
                import warnings
                warnings.warn("Angles ignored for high‑symmetry lattice ‘%s’." % lt, RuntimeWarning)

    def set_lattice_type(self, lattice_type: str, **params):
        self.lattice_type = lattice_type.lower()
        self.lattice_parameters = params

    def _set_lattice_vectors_sc(self, a: float):
        self.lattice_type = "cubic"
        self.lattice_vectors = a * np.eye(3)

    # ------------------------------------------------------------------
    #  Minimal generate_atom_positions to satisfy legacy API (origin only)
    # ------------------------------------------------------------------
    #def generate_atom_positions(self, atom_label: str = "X"):
    #    basis_frac = np.array([[0.0, 0.0, 0.0]])
    #    self._primitive_basis_cart = basis_frac @ self.lattice_vectors
    #    return self._primitive_basis_cart

    # ------------------------------------------------------------------
    #  Container helpers
    # ------------------------------------------------------------------
    def _select_sites_cartesian(self, sites_frac: np.ndarray, k: int) -> np.ndarray:
        if k > len(sites_frac):
            reps = math.ceil(k / len(sites_frac))
            sites_frac = np.tile(sites_frac, (reps, 1))[:k]
        else:
            sites_frac = sites_frac[:k]
        return sites_frac @ self.lattice_vectors

    def _populate_container(self, labels: Sequence[str], positions_cart: np.ndarray):
        container = self.add_empty_container()
        container.AtomPositionManager.set_latticeVectors(self.lattice_vectors)
        container.AtomPositionManager.add_atom(labels, positions_cart)
        #self.add_container(container)
        return container
